
const {
    JRTCClient, 
    JRTCClientInitParam, 
    JRTCClientLoginParam,
    JRTCClientCallback,
    ClientState,
    JRTCMediaDevice,
    RenderType,
    JRTCMediaDeviceVideoCanvas,
    JRTCRoom, 
    JRTCRoomInitParam, 
    JRTCRoomJoinParam,
    JRTCCallCenterCallParam,
    JRTCRecordRemoteParam,
    StompSocketAuthType,
    RoomState,
    JRTCRoomParticipant,
    RemoteRecordState,
    JRTCVersion,
    JRTCTracking,
    StorageManager,
    IntelligentMergeMode,
    VideoMergeMode,
    ScsMergeMode,
    JRTCLog,
    VideoDefinition,
    NetStatus,
    MediaType,
    JRTCMediaDeviceInitParam,
    GuestCallStateChangeType,
    JRTCGuest,
    JRTCCallCenterInitParam,
} = JRTCSDK;

function openGuest() {
    window.open("../index.html", '_self');
}

function openRoom() {
    window.open("../room/index.html", '_self');
}


const app = Vue.createApp({
    data() {
      return {
        deviceName: 'juphoon_H5',
        webSocketUrl: 'ws://192.168.2.50:8900/ws',

        client: undefined,
        mediaDevice: undefined,
        room: undefined,
        guest: undefined,
        agent: undefined,

        webSocket: undefined,
        connected: false,
        loseHeartCount: 0,
        reconnectingCount: 0,
        didClose: false,

        toastList: [],
        storage: undefined,

        HEARTBEAT_TIME: 10000, // 心跳间隔，单位ms
        HEARTBEAT_MAX_COUNT: 3, // 最大心跳丢失次数
        RECONNECT_MAX_COUNT: 6, // 最大重连次数
        keepAliveIntervalTimer: undefined,//定时器ID

        version: '',
        log: '',
        testCount: 0, // 初始化次数计数

        mediaStreamMap: undefined,
        streamUrlToUserIdMap: undefined,

        localCanvas: undefined,
        canvasMap: undefined,

        mergeParticipant: undefined
      };
    },
    created() {
        this.version = JRTCVersion.getSDKVersion();
        this.storage = new StorageManager();
        const settingFormData = this.storage.get('autoTest');
        if (settingFormData) {
            Object.entries(settingFormData)
                .forEach((entity) => this[entity[0]] = entity[1]);
        }

        window.addEventListener('beforeunload', (ev) => {
            const {webSocketUrl, deviceName} = this;
            this.storage.set('autoTest', {
                webSocketUrl,
                deviceName
            });
            ev.preventDefault();
            return '';
        });

        window.addEventListener('pagehide', (ev) => { //移动端
            const {webSocketUrl, deviceName} = this;
            this.storage.set('autoTest', {
                webSocketUrl, 
                deviceName
            });

            ev.preventDefault();
            return '';
        });
    },

    methods: {
        trust() {
            window.open(this.webSocketUrl, '_blank');
        },

        connectWebSocket() {
            console.trace("connectWebSocket");
            if ("WebSocket" in window) {
                if (this.webSocket) {
                    this.updateLog('已经存在连接...');
                    return;
                }

                if (this.webSocketUrl.trim().length  == 0) { 
                    this.showToast('自动化测试服务地址不能为空');
                    return;
                }
    
                if (this.deviceName.trim().length  == 0) { 
                    this.showToast('设备名称不能为空');
                    return;
                }
    
                const socket = new WebSocket(this.webSocketUrl);
                this.webSocket = socket;
                socket.addEventListener('open', (event) => {
                    console.trace("onOpen");
                    this.connected = true;
                    this.reconnectingCount = 0;
                    //连接成功...
                    this.updateLog("连接成功...", false);
                    this.sendMessage(this.createDeviceAddJsonString());
                    this.startKeepAlive();
                });
    
                socket.addEventListener('message', (event) => {
                    //收到消息...
                    this.updateLog("收到消息：\n" + event.data);
                    this.handleReceivedMessage(event.data);
                });
    
                socket.addEventListener('close', (event) => {
                    console.trace("onClose");
                    this.webSocket = undefined;
                    this.connected = false;
                    this.stopKeepAlive();
                    if (this.didClose) {
                        this.reconnectingCount = 0;
                        this.didClose = false;
                        this.updateLog("连接关闭... 错误码 -> " + event.code + " 原因 -> " + event.reason);
                    } else {
                        if (event.code === 4001) {
                            //心跳超时断开重连
                        } else {
                            this.updateLog("连接关闭... 错误码 -> " + event.code + " 原因 -> " + event.reason);
                        }
                        this.reconnect();  //重连
                    }
                });
    
                socket.addEventListener('error', (event) => {
                    //this.connected = false;
                    //this.stopKeepAlive();
                    console.error("WebSocket error observed:", event);
                    this.updateLog("连接失败... " + event);
                    //this.webSocket = undefined;
                    if (!this.didClose) {
                        //this.reconnect();
                    } else {
                        //this.reconnectingCount = 0;
                        //this.didClose = false;
                    }
                });
            }
        },

        reconnect() {
            console.trace("reconnect");
            this.reconnectingCount ++;
            if (this.reconnectingCount > this.RECONNECT_MAX_COUNT) {
                this.updateLog("重连超过最大" + this.RECONNECT_MAX_COUNT + "次，结束重连");
                this.reconnectingCount = 0;
            } else {
                this.updateLog("开始第" + this.reconnectingCount + "次重连");
                if (!this.didClose) {//如果主动断开连接就不用再重连了
                    this.connectWebSocket();
                } else {
                    this.didClose = false;
                    this.updateLog("重连终止... 原因 -> 主动断开连接");
                }
            }
        },

        disConnectWebSocket(code, reason) {
            console.trace("disConnectWebSocket: ", code, reason);
            this.updateLog(reason);
            this.stopKeepAlive();
            if (this.webSocket) {
                this.sendMessage(this.createDeviceRemoveJsonString());
                //延时关闭保证移除设备消息发送成功
                let that = this;
                setTimeout(() => {
                    if (code == 4000) {
                        that.didClose = true;
                    }
                    if (that.webSocket) {
                        that.webSocket.close(code, reason);
                        that.webSocket = undefined;
                    }
                }, 500);
            } else {
                if (code == 4001) {//重连
                    this.reconnect();
                } else {
                    this.updateLog("未存在连接...");
                }
            }
        },

        startKeepAlive() {
            console.trace("startKeepAlive");
            if (!this.keepAliveIntervalTimer) {
                this.updateLog("开始心跳");
                this.loseHeartCount = 0;
                this.keepAliveIntervalTimer = setInterval(this.sendHeartbeat, 10000);
                this.sendHeartbeat();
            }
        },

        sendHeartbeat() {
            console.trace("sendHeartbeat");
            this.loseHeartCount ++;
            if (this.loseHeartCount > this.HEARTBEAT_MAX_COUNT) {
                console.error("心跳超时");
                this.disConnectWebSocket(4001, "心跳超时断开");
            } else {
                console.log("发送心跳包");
                this.updateLog("发送心跳包");
                this.sendMessage(window.JSON.stringify({"type": "request", "name": "heartbeat"}));
            }
        },

        stopKeepAlive() {
            console.trace("stopKeepAlive");
            this.loseHeartCount = 0;
            if (this.keepAliveIntervalTimer) {
                this.updateLog("停止心跳");
                clearInterval(this.keepAliveIntervalTimer);
                this.keepAliveIntervalTimer = undefined;
            }
        },

        dealKeepAliveMessage(data) {
            console.trace("dealKeepAliveMessage: ", data);
            if (data && data.result && data.result.boolValue === true) {
                //心跳成功
                this.loseHeartCount = 0;
                let showText = data.description;
                if (this.isEmpty(showText)) {
                    showText = "心跳成功";
                }
                this.updateLog(showText);
            }
        },

        updateLog(text) {
            if (text && text.trim().length > 0) {
                this.log += text;
                this.log += '\n';
                //自动滚动到底部
                this.$nextTick(()=>{
                    let scrollElem = this.$refs.scrollDiv;
                    scrollElem.scrollTop = scrollElem.scrollHeight;
                   })
            }
        },

        clearLog() {
            this.log = '';
        },

        sendMessage(text) {
            console.trace("sendMessage: ", text);
            if (this.webSocket && this.webSocket.readyState == WebSocket.OPEN) {
                this.webSocket.send(text);
                this.updateLog("发送消息：\n" + text);
            } else {
                console.error("sendMessage fail webSocket not open");
            }
        },

        handleReceivedMessage(text) {
            let data = window.JSON.parse(text);
            let name = data.name;
            let type = data.type;
            
            console.trace("handleReceivedMessage: ", data);
            if ("request" === type) {
                let description = data.description;
                let dm = data.dm;
                let parameters = data.parameters;
                let returnValue = undefined;
                let async = false;
                switch (dm) {
                    case "JRTCClient":
                        switch (name) {
                            case "create":
                                // 清除console缓存
                                if (this.testCount > 100) {
                                    this.testCount = 0;
                                    this.clearLog();
                                    console.clear();
                                }
                                this.testCount ++;

                                const clientInitParam = new JRTCClientInitParam();
                                clientInitParam.appName = parameters.initParam.appName;
                                clientInitParam.server = parameters.initParam.server;
                                clientInitParam.appKey = parameters.initParam.appKey;
                                clientInitParam.logConsole = parameters.initParam.logConsole;
                                this.client = JRTCClient.create({
                                    onLogin: (result, reason) => {
                                        this.sendMessage(this.createResponseJsonString("onLogin", "登录结果回调", 
                                        dm, {result: result, reason: reason}));
                                    },
                                    onLogout: (reason) => {
                                        this.sendMessage(this.createResponseJsonString("onLogout", "登出结果回调", 
                                        dm, {reason: reason}));
                                    },
                                    onClientStateChanged: (state, oldState) => {
                                        this.sendMessage(this.createResponseJsonString("onClientStateChanged", "登录状态变化通知", 
                                        dm, {state: state, oldState: oldState}));
                                    },
                                    onSDKEvent: (event) => {
                                        this.sendMessage(this.createResponseJsonString("onSDKEvent", "上报事件回调", 
                                        dm, {event: event}));
                                    },
                                    onOnlineMessageReceived: (message, userId) => {
                                        this.sendMessage(this.createResponseJsonString("onOnlineMessageReceived", "收到在线消息回调", 
                                        dm, {message: message, userId: userId}));
                                    },
                                    onOnlineMessageSendResult: (result, operatorId) => {
                                        this.sendMessage(this.createResponseJsonString("onOnlineMessageSendResult", "在线消息发送结果回调", 
                                        dm, {result: result, operatorId: operatorId}));
                                    }
                                }, clientInitParam);
                                description = "创建 JRTCClient 对象" + (this.notUAN(this.client) ? "成功" : "失败" );
                                returnValue= this.notUAN(this.client);
                                break;
                            case "destroy":
                                if (this.notUAN(this.client)) {
                                    JRTCClient.destroy();
                                    this.client = undefined;
                                }
                                returnValue = true;
                                break;
                            case "getUserId":
                                returnValue = this.client.getUserId();
                                break;
                            case "getState":
                                returnValue = this.client.getState();
                                break;
                            case "getDisplayName":
                                returnValue = this.client.getDisplayName();
                                break;
                            case "setDisplayName":
                                this.client.setDisplayName(parameters.displayName);
                                returnValue = true;
                            case "getServer":
                                returnValue = this.client.getServer();
                                break;
                            case "setServer":
                                this.client.setServer(parameters.server);
                                returnValue = true;
                                break;
                            case "getAppKey":
                                returnValue = this.client.getAppKey();
                                break;
                            case "setAppKey":
                                this.client.setAppKey(parameters.appKey);
                                returnValue = true;
                                break;
                            case "getAppName":
                                returnValue = this.client.getAppName();
                                break;
                            case "setAppName":
                                this.client.setAppName(parameters.appName);
                                returnValue = true;
                                break;
                            case "login":
                                let loginParam = new JRTCClientLoginParam();
                                if (parameters.clientLoginParam) {
                                    loginParam.token = parameters.clientLoginParam.token;
                                    loginParam.tokenType = parameters.clientLoginParam.tokenType;
                                }
                                returnValue = this.client.login(parameters.userId, parameters.password, loginParam);
                                break;
                            case "logout":
                                returnValue = this.client.logout();
                                break;
                            case "sendOnlineMessage":
                                returnValue = this.client.sendOnlineMessage(parameters.message, parameters.userId);
                                break;
                            default:
                                description = "未找到该方法";
                                returnValue = false;
                                break;
                        }
                        break;
                    case "JRTCMediaDevice":
                        switch (name) {
                            case "create":
                                this.canvasMap = new Map();
                                let mediaDeviceInitParam = new JRTCMediaDeviceInitParam();
                                this.mediaDevice = JRTCMediaDevice.create({
                                    onVolumeChanged: (volume, userId)=> {
                                        this.sendMessage(this.createResponseJsonString("onVolumeChanged", "媒体音量上报", 
                                        dm, {volume: volume, userId: userId}));
                                    },
                                    onCameraUpdate: ()=> {
                                        this.sendMessage(this.createResponseJsonString("onCameraUpdate", "摄像头变化回调", 
                                        dm, {result: true}));
                                    },
                                    onAudioInputUpdate: ()=> {
                                        this.sendMessage(this.createResponseJsonString("onAudioInputUpdate", "输入音频变化回调", 
                                        dm, true));
                                    }
                                }, mediaDeviceInitParam);
                                returnValue = this.notUAN(this.mediaDevice);
                                break;
                            case "destroy":
                                if (this.notUAN(this.mediaDevice)) {
                                    this.mediaDevice.stopAllMedia();
                                    this.canvasMap = undefined;
                                    this.clearStreamNode();
                                    JRTCMediaDevice.destroy();
                                    this.mediaDevice = undefined;
                                }
                                returnValue = true;
                                break;
                            case "getAudioInputs":
                                returnValue = this.mediaDevice.getAudioInputs();
                                break;
                            case "getAudioOutputs":
                                returnValue = this.mediaDevice.getAudioOutputs();
                                break;
                            case "setCameraProperty":
                                this.mediaDevice.setCameraProperty(parameters.width, parameters.height, parameters.frameRate);
                                returnValue = true;
                                break;
                            case "startCameraVideo": {
                                let viewId = 'local-video';
                                this.appendLocalVideo(viewId);
                                let box = document.querySelector(`#${viewId}`);
                                this.mediaDevice.startCameraVideo(parameters.renderType, box).then((canvas) => {
                                    let ret;
                                    if (!canvas) {
                                        this.showToast(`本地视频渲染失败`);
                                        ret = undefined;
                                    } else {
                                        this.canvasMap.set(canvas.streamUrl, canvas);
                                        this.localCanvas = canvas;
                                        ret = canvas.streamUrl;
                                    }
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, ret));
                                }).catch(err => {
                                    console.error("startCameraVideo ", err);
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, undefined));
                                });
                                async = true;
                                break;
                            }  
                            case "startVideo": {
                                let viewId = 'video-';
                                if (parameters.streamUrl && this.streamUrlToUserIdMap.has(parameters.streamUrl)) {
                                    let userId = this.streamUrlToUserIdMap.get(parameters.streamUrl);
                                    viewId = 'video-' + userId;
                                    this.appendRemoteVideo(userId, false);
                                    let box = document.getElementById(`${viewId}`);
                                    let mediaStream = this.mediaStreamMap.get(parameters.mediaStreamId);
                                    const canvas = this.mediaDevice.startVideo(parameters.streamUrl, parameters.renderType, mediaStream, box);
                                    if (!canvas) {
                                        returnValue = undefined;
                                    } else {
                                        this.canvasMap.set(canvas.streamUrl, canvas);
                                        returnValue = canvas.streamUrl;
                                    }
                                } else {
                                    returnValue = undefined;
                                }
                                break;
                            }
                            case "stopVideo":
                                if (parameters.streamUrl && this.canvasMap.has(parameters.streamUrl)) {
                                    this.mediaDevice.stopVideo(this.canvasMap.get(parameters.streamUrl));
                                    this.canvasMap.delete(parameters.streamUrl);
                                    if (this.localCanvas && this.localCanvas.streamUrl === parameters.streamUrl) {
                                        this.localCanvas = undefined;
                                        this.removeLocalVideo();
                                    } else {
                                        let userId = this.streamUrlToUserIdMap.get(parameters.streamUrl);
                                        if (userId) {
                                            this.removeRemoteVideo(userId);
                                        }
                                    }
                                    returnValue = true;
                                } else {
                                    returnValue = false;
                                }
                                break;
                            case "stopAllMedia":
                                this.mediaDevice.stopAllMedia();
                                this.canvasMap.clear();
                                this.clearStreamNode();
                                returnValue = true;
                                break;
                            case "startAudioInput":
                                this.mediaDevice.startAudioInput().then((result) => {
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, result));
                                }).catch(err => {
                                    console.error("startAudioInput ", err);
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, false));
                                });
                                async = true;
                                break;
                            case "stopAudioInput": 
                                returnValue = this.mediaDevice.stopAudioInput();
                                break;
                            case "startAudioOutput":
                                let mediaStream = this.mediaStreamMap.get(parameters.mediaStreamId);
                                returnValue = this.mediaDevice.startAudioOutput(mediaStream, parameters.userId);
                                break;
                            case "stopAudioOutput":
                                returnValue = this.mediaDevice.stopAudioOutput(parameters.userId);
                                break;
                            case "switchCamera":
                                let camaras = this.mediaDevice.getCameras();
                                let camara = undefined;
                                for (const c in camaras) {
                                    if (parameters.cameraId === c.cameraId) {
                                        camara = c;
                                        break;
                                    }
                                }
                                if (camara) {
                                    this.mediaDevice.switchCamera(camara).then((result) => {
                                        this.sendMessage(this.createReturnJsonString(name, description, dm, result));
                                    }).catch(err => {
                                        console.error("switchCamera ", err);
                                        this.sendMessage(this.createReturnJsonString(name, description, dm, false));
                                    });
                                    async = true;
                                } else {
                                    returnValue = false;
                                }
                                break;
                            case "isCameraOpen":
                                returnValue = this.mediaDevice.isCameraOpen();
                                break;
                            case "getCurrentCamera":
                                returnValue = this.mediaDevice.getCurrentCamera();
                                break;
                            case "getCameras":
                                returnValue = this.mediaDevice.getCameras();
                                break;
                            case "isAudioStart":
                                returnValue = this.mediaDevice.isAudioStart();
                                break;
                            default:
                                description = "未找到该方法";
                                returnValue = false;
                                break;
                        }
                        break;
                    case "JRTCRoom":
                        switch (name) {
                            case "create":
                                this.mediaStreamMap = new Map();
                                this.streamUrlToUserIdMap = new Map();
                                const roomInitParam = new JRTCRoomInitParam();
                                roomInitParam.roomServer = parameters.initParam.roomServer;
                                roomInitParam.protocol = parameters.initParam.protocol;
                                this.room = JRTCRoom.create(this.client, this.mediaDevice, roomInitParam, {
                                    onRoomStateChanged: (state, oldState) => {
                                        this.sendMessage(this.createResponseJsonString("onRoomStateChanged", "自身在房间中的状态变化回调", 
                                        dm, {state: state, oldState: oldState}));
                                    },
                                    onRoomPropertyChanged: (changeParam) => {
                                        this.sendMessage(this.createResponseJsonString("onRoomPropertyChanged", "房间属性变化回调", 
                                        dm, {changeParam: changeParam}));
                                    },
                                    onJoin: (result, reason, roomId) => {
                                        this.mediaStreamMap.clear();
                                        if (this.canvasMap) {
                                            this.canvasMap.clear();
                                        }
                                        this.streamUrlToUserIdMap.clear();
                                        this.sendMessage(this.createResponseJsonString("onJoin", "加入房间结果回调", 
                                        dm, {result: result, reason: reason, roomId: roomId}));
                                    },
                                    onLeave: (reason, roomId) => {
                                        this.mediaStreamMap.clear();
                                        if (this.canvasMap) {
                                            this.canvasMap.clear();
                                        }
                                        this.streamUrlToUserIdMap.clear();
                                        this.sendMessage(this.createResponseJsonString("onLeave", "离开房间结果回调", 
                                        dm, {reason: reason, roomId: roomId}));
                                    },
                                    onQuery: (operationId, result, reason, queryInfo) => {
                                        this.sendMessage(this.createResponseJsonString("onQuery", "查询房间信息结果回调", 
                                        dm, {operationId: operationId, result: result, reason: reason, queryInfo: queryInfo}));
                                    },
                                    onParticipantJoin: (participant) => {
                                        if ("MERGED__" == participant.userId) {
                                            this.mergeParticipant = participant;//保存合流成员
                                        }
                                        this.sendMessage(this.createResponseJsonString("onParticipantJoin", "新成员加入回调", 
                                        dm, {participant: participant}));
                                    },
                                    onParticipantLeft: (participant, reason) => {
                                        if ("MERGED__" == participant.userId) {
                                            this.mergeParticipant = undefined;
                                        }
                                        this.sendMessage(this.createResponseJsonString("onParticipantLeft", "成员离开回调", 
                                        dm, {participant: participant, reason: reason}));
                                    },
                                    onParticipantUpdate: (participant, changeParam) => {
                                        this.sendMessage(this.createResponseJsonString("onParticipantUpdate", "成员更新回调", 
                                        dm, {participant: participant, changeParam: changeParam}));
                                    },
                                    onMessageReceived: (type, content, fromUserId) => {
                                        this.sendMessage(this.createResponseJsonString("onMessageReceived", "接收房间消息的回调", 
                                        dm, {type: type, content: content, fromUserId: fromUserId}));
                                    },
                                    onDeliveryAbort: (isShutDown, deliveryUserId, reason) => {
                                        this.sendMessage(this.createResponseJsonString("onDeliveryAbort", "录制异常回调", 
                                        dm, {isShutDown: isShutDown, deliveryUserId: deliveryUserId, reason: reason}));
                                    }
                                });
                                returnValue = this.notUAN(this.room);
                                break;
                            case "destroy":
                                if (this.notUAN(this.room)) {
                                    this.mediaStreamMap = undefined; 
                                    this.streamUrlToUserIdMap = undefined;
                                    JRTCRoom.destroy();
                                    this.room = undefined;
                                }
                                returnValue = true;
                                break;
                            case "enableUploadAudioStream":
                                returnValue = this.room.enableUploadAudioStream(parameters.enable);
                                break;
                            case "enableUploadVideoStream":
                                returnValue = this.room.enableUploadVideoStream(parameters.enable);
                                break;
                            case "getRoomId":
                                returnValue = this.room.getRoomId();
                                break;
                            case "getPassword":
                                returnValue = this.room.getPassword();
                                break;
                            case "getCallId":
                                returnValue = this.room.getCallId();
                                break;
                            case "getTitle":
                                returnValue = this.room.getTitle();
                                break;
                            case "getState":
                                returnValue = this.room.getState();
                                break;
                            case "getParticipants":
                                returnValue = this.room.getParticipants();
                                break;
                            case "getParticipant":
                                returnValue = this.room.getParticipant(parameters.userId);
                                break;
                            case "statistics":
                                returnValue = this.room.statistics();
                                break;
                            case "getUploadLocalAudio":
                                returnValue = this.room.getUploadLocalAudio();
                                break;
                            case "getUploadLocalVideo":
                                returnValue = this.room.getUploadLocalVideo();
                                break;
                            case "getShareStreamId":
                                returnValue = this.room.getShareStreamId();
                                break;
                            case "getShareUserId":
                                returnValue = this.room.getShareUserId();
                                break;
                            case "getRemoteRecordState":
                                returnValue = this.room.getRemoteRecordState();
                                break;
                            case "join":
                                let joinParam = new JRTCRoomJoinParam();
                                Object.keys(parameters.joinParam).forEach(key => {
                                    joinParam[key] = parameters.joinParam[key];
                                });
                                returnValue = this.room.join(parameters.roomId, joinParam);
                                break;
                            case "leave":
                                returnValue = this.room.leave();
                                break;
                            case "requestVideo": {
                                let participant = undefined;
                                if (parameters.userId === 'MERGED__') { //合流成员
                                    participant = this.mergeParticipant;
                                } else {
                                    participant = this.room.getParticipant(parameters.userId);
                                }
                                this.room.requestVideo(participant, parameters.videoSize).then((result) => {
                                    let ret;
                                    if (result) {
                                        this.mediaStreamMap.set(result.mediaStream.id, result.mediaStream);
                                        this.streamUrlToUserIdMap.set(result.streamUrl, participant.userId);
                                        ret = {streamUrl: result.streamUrl, mediaStreamId: result.mediaStream.id};
                                    } else {
                                        ret = undefined;
                                    }
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, ret));
                                }).catch(err => {
                                    console.error("requestVideo ", err);
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, undefined));
                                });
                                async = true;
                                break;
                            }
                            case "unRequestVideo":
                                returnValue = this.room.unRequestVideo(this.room.getParticipant(parameters.userId));
                                break;
                            case "requestScreenVideo":
                                this.room.requestScreenVideo(parameters.videoSize).then((result) => {
                                    let ret;
                                    if (result) {
                                        this.mediaStreamMap.set(result.mediaStream.id, result.mediaStream);
                                        this.streamUrlToUserIdMap.set(result.streamUrl, "screenShare");
                                        ret = {streamUrl: result.streamUrl, mediaStreamId: result.mediaStream.id};
                                    } else {
                                        ret = undefined;
                                    }
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, ret));
                                }).catch(err => {
                                    console.error("requestScreenVideo ", err);
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, undefined));
                                });
                                async = true;
                                break;
                            case "unRequestScreenVideo":
                                returnValue = this.room.unRequestScreenVideo();
                                break;
                            case "enableRemoteRecord":
                                returnValue = this.room.enableRemoteRecord(parameters.enable, parameters.recordParam);
                                break;
                            case "sendMessage":
                                returnValue = this.room.sendMessage(parameters.type, parameters.content, parameters.toUserId);
                                break;
                            case "setRatio":
                                returnValue = this.room.setRatio(parameters.ratio);
                                break;
                            case "isAutoRecord":
                                returnValue = this.room.isAutoRecord();
                                break;
                            default:
                                description = "未找到该方法";
                                returnValue = false;
                                break;
                        }
                        break;
                    case "JRTCGuest":
                        switch (name) {
                            case "create":
                                this.mediaStreamMap = new Map();
                                this.streamUrlToUserIdMap = new Map();
                                const guestInitParam = new JRTCCallCenterInitParam();
                                guestInitParam.roomServer = parameters.initParam.roomServer;
                                guestInitParam.protocol = parameters.initParam.protocol;
                                this.guest = JRTCGuest.create(this.client, this.mediaDevice, guestInitParam, {
                                    onGetAllGroups: (result, groups) => {
                                        this.sendMessage(this.createResponseJsonString("onGetAllGroups", "查询业务组号回调", 
                                        dm, {result: result, groups: groups}));
                                    },
                                    onCallStateChanged: (type, incomingType, inviter, reason) => {
                                        if (type === GuestCallStateChangeType.TERMED) {
                                            this.mediaStreamMap.clear();
                                            if (this.canvasMap) {
                                                this.canvasMap.clear();
                                            }
                                            this.streamUrlToUserIdMap.clear();
                                        }
                                        this.sendMessage(this.createResponseJsonString("onCallStateChanged", "通话状态改变", 
                                        dm, {type: type, incomingType: incomingType, inviter: inviter, reason: reason}));
                                    },
                                    onCallQueueCount: (count, time, agentRinging) => {
                                        this.sendMessage(this.createResponseJsonString("onCallQueueCount", "当前排队人数上报回调", 
                                        dm, {count: count, time: time, agentRinging: agentRinging}));
                                    },
                                    onCallPropertyChanged: (propChangeParam) => {
                                        this.sendMessage(this.createResponseJsonString("onCallPropertyChanged", "通话属性改变回调", 
                                        dm, {propChangeParam: propChangeParam}));
                                    },
                                    onMemberJoin: (participant) => {
                                        if ("MERGED__" == participant.userId) {
                                            this.mergeParticipant = participant;//保存合流成员
                                        }
                                        this.sendMessage(this.createResponseJsonString("onMemberJoin", "通话中有新成员加入回调", 
                                        dm, {participant: participant}));
                                    },
                                    onMemberLeave: (participant) => {
                                        if ("MERGED__" == participant.userId) {
                                            this.mergeParticipant = undefined;
                                        }
                                        this.sendMessage(this.createResponseJsonString("onMemberLeave", "通话中有新成员离开回调", 
                                        dm, {participant: participant}));
                                    },
                                    onMemberUpdate: (participant, changeParam) => {
                                        this.sendMessage(this.createResponseJsonString("onMemberUpdate", "通话中成员属性更新回调", 
                                        dm, {participant: participant, changeParam: changeParam}));
                                    },
                                    onMessageReceived: (content, contentType, messageType, fromUserId) => {
                                        this.sendMessage(this.createResponseJsonString("onMessageReceived", "收到消息回调", 
                                        dm, {content: content, contentType: contentType, messageType: messageType, fromUserId: fromUserId}));
                                    },
                                    onSignRequest: (fromUserId, extraInfo) => {
                                        this.sendMessage(this.createResponseJsonString("onSignRequest", "收到签名请求回调", 
                                        dm, {fromUserId: fromUserId, extraInfo: extraInfo}));
                                    },
                                    onDeliveryAbort: (isShutDown, deliveryUserId, reason) => {
                                        this.sendMessage(this.createResponseJsonString("onDeliveryAbort", "录制异常回调", 
                                        dm, {isShutDown: isShutDown, deliveryUserId: deliveryUserId, reason: reason}));
                                    },
                                    onNotifyMessageReceived: (notifyMessage, fromUserId) => {
                                        this.sendMessage(this.createResponseJsonString("onNotifyMessageReceived", "收到事件通知消息回调", 
                                        dm, {notifyMessage: notifyMessage, fromUserId: fromUserId}));
                                    },
                                    onUrgentResultResponse: (agree) => {
                                        this.sendMessage(this.createResponseJsonString("onUrgentResultResponse", "座席处理加急的结果回调", 
                                        dm, {agree: agree}));
                                    },
                                    onCallTypeChanged: (callType) => {
                                        this.sendMessage(this.createResponseJsonString("onCallTypeChanged", "音视频通话切换回调", 
                                        dm, {callType: callType}));
                                    },
                                    onHoldStateChanged: (hold) => {
                                        this.sendMessage(this.createResponseJsonString("onHoldStateChanged", "收到通话保持或取回的回调", 
                                        dm, {hold: hold}));
                                    },
                                    onOnewayVideoChanged: (turnOn) => {
                                        this.sendMessage(this.createResponseJsonString("onOnewayVideoChanged", "收到单向视频状态变化回调", 
                                        dm, {turnOn: turnOn}));
                                    }

                                });
                                returnValue = this.notUAN(this.guest);
                                break;
                            case "destroy":
                                if (this.notUAN(this.guest)) {
                                    this.mediaStreamMap = undefined;
                                    this.streamUrlToUserIdMap = undefined;
                                    JRTCGuest.destroy();
                                    this.guest = undefined;
                                }
                                returnValue = true;
                                break;
                            case "getCallState":
                                returnValue = this.guest.getCallState();
                                break;
                            case "isMainGuest":
                                returnValue = this.guest.isMainGuest();
                                break;
                            case "getMainAgentParticipant":
                                returnValue = this.guest.getMainAgentParticipant();
                                break;
                            case "getAgentParticipants":
                                returnValue = this.guest.getAgentParticipants();
                                break;
                            case "getSelfParticipant":
                                returnValue = this.guest.getSelfParticipant();
                                break;
                            case "getParticipants":
                                returnValue = this.guest.getParticipants();
                                break;
                            case "getSerialNumber":
                                returnValue = this.guest.getSerialNumber();
                                break;
                            case "getCallId":
                                returnValue = this.guest.getCallId();
                                break;
                            case "getShareStreamId":
                                returnValue = this.guest.getShareStreamId();
                                break;
                            case "getShareUserId":
                                returnValue = this.guest.getShareUserId();
                                break;
                            case "setMultiStream":
                                this.guest.setMultiStream(parameters.multiStream);
                                returnValue = true;
                                break;
                            case "queryAllGroups":
                                returnValue = this.guest.queryAllGroups();
                                break;
                            case "call": {
                                let callParam = new JRTCCallCenterCallParam();
                                if (parameters.callParam) {
                                    Object.keys(parameters.callParam).forEach(key => {
                                        callParam[key] = parameters.callParam[key];
                                    });
                                }
                                returnValue = this.guest.call(parameters.number, callParam);
                                break;
                            }
                            case "oneToOneCall": {
                                let callParam = new JRTCCallCenterCallParam();
                                Object.keys(parameters.callParam).forEach(key => {
                                    callParam[key] = parameters.callParam[key];
                                });
                                returnValue = this.guest.oneToOneCall(parameters.userId, callParam);
                                break;
                            }
                            case "answer":
                                returnValue = this.guest.answer();
                                break;
                            case "term":
                                returnValue = this.guest.term();
                                break;
                            case "requestUrgent":
                                returnValue = this.guest.requestUrgent();
                                break;
                            case "sendMessage":
                                returnValue = this.guest.sendMessage(parameters.contentType, parameters.content, parameters.toUserId);
                                break;
                            case "sendNotifyMessage":
                                returnValue = this.guest.sendNotifyMessage(parameters.notifyMessage, parameters.toUserId);
                                break;
                            case "enableUploadAudioStream":
                                returnValue = this.guest.enableUploadAudioStream(parameters.enable);
                                break;
                            case "enableUploadVideoStream":
                                returnValue = this.guest.enableUploadVideoStream(parameters.enable);
                                break;
                            case "setRatio":
                                returnValue = this.guest.setRatio(parameters.ratio);
                                break;
                            case "getRequestSize":
                                returnValue = this.guest.getRequestSize();
                                break;
                            case "setRequestSize":
                                returnValue = this.guest.setRequestSize(parameters.requestSize);
                                break;
                            case "requestVideo": {
                                let participant = undefined;
                                if (parameters.userId === 'MERGED__') { //合流成员
                                    participant = this.mergeParticipant;
                                } else {
                                    participant = this.getParticipant(this.guest.getParticipants(), parameters.userId);
                                }
                                this.guest.requestVideo(participant, parameters.videoSize).then((result) => {
                                    let returnValue;
                                    if (result) {
                                        this.mediaStreamMap.set(result.mediaStream.id, result.mediaStream);
                                        this.streamUrlToUserIdMap.set(result.streamUrl, participant.userId);
                                        returnValue = {streamUrl: result.streamUrl, mediaStreamId: result.mediaStream.id};
                                    } else {
                                        returnValue = undefined;
                                    }
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, returnValue));
                                }).catch(err => {
                                    console.error("requestVideo ", err);
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, undefined));
                                });
                                async = true;
                                break;
                            }
                            case "unRequestVideo":
                                returnValue = this.guest.unRequestVideo(this.getParticipant(this.guest.getParticipants(), parameters.userId));
                                break;
                            case "requestScreenVideo":
                                this.guest.requestScreenVideo(parameters.videoSize).then((result) => {
                                    let returnValue;
                                    if (result) {
                                        this.mediaStreamMap.set(result.mediaStream.id, result.mediaStream);
                                        this.streamUrlToUserIdMap.set(result.streamUrl, "screenShare");
                                        returnValue = {streamUrl: result.streamUrl, mediaStreamId: result.mediaStream.id};
                                    } else {
                                        returnValue = undefined;
                                    }
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, returnValue));
                                }).catch(err => {
                                    console.error("requestScreenVideo ", err);
                                    this.sendMessage(this.createReturnJsonString(name, description, dm, undefined));
                                });
                                async = true;
                                break;
                            case "unRequestScreenVideo":
                                returnValue = this.guest.unRequestScreenVideo();
                                break;
                            case "enableRemoteRecord":
                                returnValue = this.guest.enableRemoteRecord(parameters.enable, parameters.recordParam);
                                break;
                            case "getRemoteRecordState":
                                returnValue = this.guest.getRemoteRecordState();
                                break;
                            case "responseSignResult":
                                returnValue = this.guest.responseSignResult(parameters.result, parameters.userId, parameters.extraInfo);
                                break;
                            case "statistics":
                                returnValue = this.guest.statistics();
                                break;
                            default:
                                description = "未找到该方法";
                                returnValue = false;
                                break;
                        }
                        break;
                    default:
                        switch (name) {
                            case "stopTest":
                                if (this.guest) {
                                    this.mediaStreamMap = undefined;
                                    this.streamUrlToUserIdMap = undefined;
                                    JRTCGuest.destroy();
                                    this.guest = undefined;
                                }
                                if (this.room) {
                                    this.mediaStreamMap = undefined;
                                    this.streamUrlToUserIdMap = undefined;
                                    JRTCRoom.destroy();
                                    this.room = undefined;
                                }
                                if (this.mediaDevice) {
                                    this.mediaDevice.stopAllMedia();
                                    this.canvasMap = undefined;
                                    this.clearStreamNode();
                                    JRTCMediaDevice.destroy();
                                    this.mediaDevice = undefined;
                                }
                                if (this.client) {
                                    JRTCClient.destroy();
                                    this.client = undefined;
                                }
                                returnValue = true;
                                break;
                            default:
                                description = "未找到该方法";
                                returnValue = false;
                                break;
                        }
                        break;
                }
                if (!async) {
                    this.sendMessage(this.createReturnJsonString(name, description, dm, returnValue));
                }
            } else if ("response" === type) {
                if ("heartbeat" === name) {
                    this.dealKeepAliveMessage(data);
                }
            }
        },

        getParticipant(participants, userId) {
            for (const p of participants) {
                if (p.userId === userId) {
                    return p;
                }
            }
            return undefined;
        },

        createReturnJsonString(name, description, dm, result) {
            let data = this.createCommonJsonObj('return', name, description);
            data.dm = dm;
            data.result = result;
            return window.JSON.stringify(data);
        },

        createResponseJsonString(name, description, dm, result) {
            let data = this.createCommonJsonObj('response', name, description);
            data.dm = dm;
            data.result = result;
            return window.JSON.stringify(data);
        },

        createCommonJsonObj(type, name, description) {
            let data = {
                type: type,
                name: name,
                description: description,
                time: new Date().getTime()
            }
            return data;
        },

        createDeviceAddJsonString() {
            let data = this.createCommonJsonObj('request', 'deviceAdd', '添加设备');
            data.system = 'H5';
            data.nickname = this.deviceName;
            data.version = this.version;
            data.deviceId = this.uuid();
            return window.JSON.stringify(data);
        },

        createDeviceRemoveJsonString() {
            let data = this.createCommonJsonObj('request', 'deviceRemove', '移除设备');
            data.system = 'H5';
            data.nickname = this.deviceName;
            return window.JSON.stringify(data);
        },

        uuid() {
            let s = [];
            let hexDigits = "0123456789abcdef";
            for (let i = 0; i < 36; i++) {
                let from = Math.floor(Math.random() * 0x10);
                s[i] = hexDigits.substring(from, from + 1);
            }
            s[14] = "4"; // bits 12-15 of the time_hi_and_version field to 0010
            let from = (s[19] & 0x3) | 0x8;
            s[19] = hexDigits.substring(from, from + 1); // bits 6-7 of the clock_seq_hi_and_reserved to 01
            s[8] = s[13] = s[18] = s[23] = "-";
         
            let uuid = s.join("");
            return uuid;
        },

        notUAN(object) {
            if (object != null && object != undefined) {
                return true;
            } else {
                return false;
            }
        },

        isEmpty(str) {
            if (
              str === null ||
                  str === '' ||
                  str === undefined ||
                  str.length === 0 || str.trim().length === 0
            ) {
              return true
            } else {
              return false
            }
        },

        isNotEmpty(str) {
            if (
                str === null ||
                    str === '' ||
                    str === undefined ||
                    str.length === 0 || str.trim().length === 0
              ) {
                return false
              } else {
                return true
              }
        },

        
        /**
         * helper function
         */
        showToast(msg, duration) {
            this.toastList.push(msg);

            let length = this.toastList.length;
            let top = 5 * length  + 30;

            duration = isNaN(duration) ? 3000 : duration;
            var m = document.createElement('div');
            m.innerHTML = msg;
            m.style.cssText = `width:50%; min-width:180px; background:#000; opacity:0.5; height:auto;min-height: 30px; color:#fff; line-height:30px; text-align:center; border-radius:4px; position:fixed; top:${top}%; left:20%; z-index:999999;`
            document.body.appendChild(m);

            let that = this;
            setTimeout(function() {
                var d = 0.5;
                m.style.webkitTransition = '-webkit-transform ' + d + 's ease-in, opacity ' + d + 's ease-in';
                m.style.opacity = '0';

                setTimeout(function() {
                        document.body.removeChild(m);
                        that.toastList.splice(0, 1);
                    }, d * 1000);
            }, duration);
        },

        isScreenShareParticipant(participant) {
            if (participant.userId === this.room.getCallId()) {
             return true;
            }
            return false;
        },

        clearStreamNode() {
            const remote = document.getElementById("remote-video-container");
            if (remote != null) {
              this.clearAllNode(remote);
            }
            
            const local = document.getElementById("local-video-container");
            if (local != null) {
              this.clearAllNode(local);
            }
        },
    
        clearAllNode(parentNode){
            parentNode.innerHTML = '';
        },

        appendErrorContent(content) {
            let div =  this.createDivDOM();
            div.textContent = content;
            let viewId = 'error-code-detail'
            const box = document.querySelector(`#${viewId}`);
            box.appendChild(div);
        },

        createDivDOM() {
            const divDom = document.createElement('div');
            divDom.style.display = 'flex';
            divDom.style.margin = '0 auto';
            divDom.style.fontSize = '10px';
            return divDom;
        },

        appendLocalVideo(vid) {

            let userId = this.client.getUserId();

            let viewId = 'local-video-container'
            let local_video_container = document.querySelector(`#${viewId}`);
            const figure = document.createElement('figure');
            figure.id = 'figure-' + userId;

            let div = this.createDivDOM();
            div.style.width = '180px';
            div.style.height = '320px';
            div.id = vid;
            div.style.backgroundColor = 'rgba(0,0,0)';
            figure.appendChild(div);

            let textDiv = this.createDivDOM();
            textDiv.style.width = '180px';
            textDiv.textContent = 'self';
            textDiv.style.display = 'flex';
            textDiv.style.justifyContent = 'center';
            figure.appendChild(textDiv);

            local_video_container.appendChild(figure);
        },

        removeLocalVideo() {
            let userId = this.client.getUserId();

            let viewId = 'local-video-container';
            let local_video_container = document.querySelector(`#${viewId}`);

            let videoViewId = 'figure-' + userId;
            let video = document.getElementById(`${videoViewId}`);

            if (video) {
                local_video_container.removeChild(video);
            }
        },

        appendRemoteVideo(userId, isScreenShare) {

            let viewId = 'remote-video-container'
            let remote_video_container = document.querySelector(`#${viewId}`);
            const figure = document.createElement('figure');
            
            figure.className = 'stream-wrapper';
            figure.id = 'figure-' + userId;

            let div = this.createDivDOM();
            div.style.width = '180px';
            div.style.height = '320px';
            div.id = 'video-' + userId;
            div.style.backgroundColor = 'rgba(0,0,0)';
            figure.appendChild(div);

            let textDiv = this.createDivDOM();
            textDiv.style.width = '180px';
            textDiv.textContent = userId;
            textDiv.style.display = 'flex';
            textDiv.style.justifyContent = 'center';
            figure.appendChild(textDiv);

            remote_video_container.appendChild(figure);
        },

        removeRemoteVideo(userId) {
            let viewId = 'remote-video-container';
            let remote_video_container = document.querySelector(`#${viewId}`);

            let videoViewId = 'figure-' + userId;
            let video = document.getElementById(`${videoViewId}`);
            if (video) {
                remote_video_container.removeChild(video);
            }    
        },

        openFAQPage() {
            window.open(`../FAQ.html`);
        },

        strSize(str, charset) {
            let total = 0;
            charset = charset?.toLowerCase() || '';
            for (i = 0; i < str.length; i++) {
                let charCode = str.charCodeAt(i);
                if (charset === 'utf-16' || charset === 'utf16') {
                    total += charCode <= 0xffff ? 2 : 4;
                } else {
                    if (charCode <= 0x007f) {
                        total += 1;
                    } else if (charCode <= 0x07ff) {
                        total += 2;
                    } else if (charCode <= 0xffff) {
                        total += 3;
                    } else {
                        total += 4;
                    }
                }
            }
            return total;
        }
    }
  }).mount('#app');