const {
    JRTCClient,
    JRTCClientInitParam,
    JRTCClientLoginParam,
    JRTCClientCallback,
    ClientState,
    JRTCMediaDevice,
    RenderType,
    JRTCMediaDeviceVideoCanvas,
    JRTCRoom,
    JRTCRoomInitParam,
    JRTCRoomJoinParam,
    StompSocketAuthType,
    RoomState,
    JRTCRoomParticipant,
    RemoteRecordState,
    JRTCVersion,
    JRTCGuest,
    JRTCCallCenterCallParam,
    JRTCCallCenterInitParam,
    GuestCallStateChangeType,
    JRTCTracking,
    CallState,
    CallType,
    StorageManager,
    JRTCLog,
    JRTCLogLevel,
    VideoDefinition,
    NetStatus
} = JRTCSDK;

function openRoom() {
    window.open("./room/index.html", '_self');
}

function openAutoTest() {
    window.open("./autotest/index.html", '_self');
}

const app = Vue.createApp({
    data() {
      return {

        //client的环境参数
        clientEnv: {
            server: 'https://192.168.4.171:10034',
            appKey: '05c97ce0d0a84d2592334093',
            token: '',
            tokenType: 'jrtc_access'
        },

        //房间的环境参数
        roomEnv: {
            server: 'https://192.168.4.171:10034',
        },

        /**
         * client
         */
        client: null,

        accountSettings: {
            userId: '',
            displayName: ''
        },

        clienstState: ClientState.NOT_INIT,

        isInitialized: false,

        clientStateText : '未登录',

        isLogin: false,

        version: '',

        /**
         * mediaDevice
         */
        mediaDevice: null,

        cameraList: [],

        microphoneList: [],
        
        localCanvas : null,


        // 音量progress map
        volumeMap: {},

        renderTypeMap: [
            {'label': 'FullContent', value: RenderType.RENDER_FULL_CONTENT},
            {'label': 'FullScreen', value: RenderType.RENDER_FULL_SCREEN},
        ],

        // 渲染模式
        renderType: RenderType.RENDER_FULL_CONTENT,

        /**
         * guest
         */
        guest: null,

        guestCallback: null,

        maxResolutionMap: [
            {'label': '360p', value: 0},
            {'label': '720p', value: 1},
            {'label': '1080p', value: 2},
        ],

        securityTypeMap: [
            {'label': '不加密', value: 0},
            {'label': 'SRTP', value: 1},
            {'label': 'SM4', value: 2},
        ],

        videoEncodeTypeMap: [
            {'label': 'H264', value: 0},
            {'label': 'AV1', value: 1},
        ],

        audioEncodeTypeMap: [
            {'label': 'OPUS', value: 0},
            {'label': 'PCMA', value: 1},
            {'label': 'PCMU', value: 2},
        ],

        onlineMessage: {
            content: '',
            userId: '',
        },

        mediaSettings: {
            isSelfDefined: false,
            maxResolution: 0,
            svcResolution: '1 90 70 180 180 360 480 720 1350',
            captureResolution: 1,
            frameRate: 24
        },

        roomSettings: {
            currentCamera: null,
            currentMic: null,
            sendMessage: {
                type: '',
                content: '',
                userId: '',
            },
            participants: [],
        },

        callAgentSettings: {
            telNumber: '',
            agentUserId: '',
            displayName: '',

            autoRecord: true,
            maxResolution: 0,
            svcResolution: "1 90 70 180 180 360 480 720 1350",

            heartbeatTime: 20,
            heartbeatTimeout: 60,

            videoEncodeType: 0,
            audioEncodeType: 0,
            extraInfo: '',
            securityType: 0,
            isMultiStream: false,

            captureWidth: 360,
            captureHeight: 640,
            frameRate: 24,
            videoDefinition: 0, // VideoDefinition 
            vipPower: 0, // 呼叫权重
            traceId: '' //业务流水号
        },

        queueInfo: '',

        publishingVideo: false, // 是否正在视频推流
        publishingAudio: false, // 是否正在音频推流

        callState: 'idle',

        callStateText: '空闲',

        isTalking: false,

        // 是否已经加速
        isUrgent: false,

        screenShareCanvas: null,

        /**
         * toast
         */

         toastList: [],
         /**
         * Stats
         */
        statsInfo: '',
        statsInternalKey: null,

        // 日志存储
        logStore: true

      };
    },
    created() {
        this.version = JRTCVersion.getSDKVersion();

        this.storage = new StorageManager();

        const settingFormData = this.storage.get('settingFormData_webrtc-guest-sdk');
        if (settingFormData) {
        Object.entries(settingFormData)
            .forEach((entity) => this[entity[0]] = entity[1]);
        }
        if (!this.accountSettings) {
            this.accountSettings = {};
        }

        window.addEventListener('beforeunload', (ev) => {
            const {roomSettings, callAgentSettings, mediaSettings, clientEnv, roomEnv, accountSettings, logStore} = this;

            this.storage.set('settingFormData_webrtc-guest-sdk', {
                roomSettings,
                callAgentSettings,
                mediaSettings,
                clientEnv,
                roomEnv,
                accountSettings,
                logStore
            });

            ev.preventDefault();
            ev.returnValue = '';
            return '';
        });

        window.addEventListener('pagehide', (ev) => {
            const {roomSettings, callAgentSettings, mediaSettings, clientEnv, roomEnv, accountSettings, logStore} = this;

            this.storage.set('settingFormData_webrtc-guest-sdk', {
                roomSettings,
                callAgentSettings,
                mediaSettings,
                clientEnv,
                roomEnv,
                accountSettings,
                logStore
            });

            ev.preventDefault();
            ev.returnValue = '';
            return '';
        });
        
    },
    methods: {

        trust() {
            window.open(this.clientEnv.server, '_blank');
        },

        /**
         * client
         */
        clientInit() {
            const clientInitParam = new JRTCClientInitParam();
            clientInitParam.appName = "JCCSample-H5";
            clientInitParam.server = this.clientEnv.server;
            clientInitParam.appKey = this.clientEnv.appKey;
            this.clienstState = ClientState.IDLE;
            this.client = JRTCClient.create({
                onLogin: (result, reason) => {
                    console.log('onLogin result: ', result, ' reason: ', reason);
                },
                onLogout: (reason) => {
                    console.log('onLogout reason: ', reason);
                    this.isUrgent = false;
                    this.callState = 'idle';
                    this.term();
                },
                onClientStateChanged: (state, oldState) => {
                    console.log('onClientStateChanged state: ', state);
                    this.clienstState = state;
                    this.isLogin = state == ClientState.LOGINED || state == ClientState.LOGINING;

                    this.onClientStatusChanged(state);

                    this.callState = this.isLogin ? 'idle' :  this.callState;
                },
                onSDKEvent: (event)=> {
                    console.log('onSDKEvent event: ', event);
                    if (event.errorCode && event.errorCode.length > 0 && event.errorCodeDetail && event.errorCodeDetail.length > 0) {
                        let content = "错误码:" + event.errorCode + "错误描述:" + event.errorCodeDetail;
                        this.appendErrorContent(content);
                    }
                },
                onOnlineMessageReceived: (message, userId) => {
                    this.showToast('收到来自 ' + userId + ' 发送的在线消息: ' + message);
                },
                onOnlineMessageSendResult: (result, operatorId) => {
                    let content = '发送';
                    content += result ? '成功' : '失败';
                    content += ' operatorId: ' + operatorId;
                    console.log(content);
                }
            }, clientInitParam);
            this.client.setDisplayName(this.callAgentSettings.displayName);
            JRTCTracking.setChannel("H5 channel");

            JRTCLog.localLogStore = this.logStore;
            JRTCLog.setLevel(JRTCLogLevel.LOG_LEVEL_TRACE);
        },

        initHandle() {
            if (this.isInitialized) {
                if (this.isLogin) {
                    this.client.logout();
                }
                JRTCClient.destroy();
                this.client = null;

                JRTCMediaDevice.destroy();
                this.mediaDevice = null;

                if (this.isTalking) {
                    this.guest.term();
                }
                JRTCGuest.destroy();
                this.guest = null;
            } else {
                this.clientInit();
                this.mediaDeviceInit();
                this.guestInit();
            }
            this.isInitialized = !this.isInitialized;
        },


        downloadLog() {
            
            const blob = JRTCLog.getLogMessage(),
            blobUrl = window.URL.createObjectURL(blob),
            tempLink = document.createElement('a');

            tempLink.style.display = 'none';
            tempLink.href = blobUrl;
            tempLink.setAttribute('download', `H5 ${window.location.host}-${Date.now()}.log`);

            if (typeof tempLink.download === 'undefined') {
                tempLink.setAttribute('target', '_blank');
            }

            document.body.appendChild(tempLink);
            tempLink.click();
            document.body.removeChild(tempLink);

            setTimeout(() => {
                window.URL.revokeObjectURL(blobUrl);
            });
        },

        clearLog() {
            JRTCLog.clearLogMessage();
        },

        loginHandle() {
            if (this.isLogin) {
                this.client.logout();
                if (this.callState != 'idle' && this.callState != 'termed') {
                    this.guest.term();
                }
            } else {
                if (!this.client) {
                    this.clientInit();
                }
                let loginParam = new JRTCClientLoginParam();
                loginParam.token = this.clientEnv.token;
                loginParam.tokenType = this.clientEnv.tokenType;
                this.client.setServer(this.clientEnv.server);
                this.client.setAppKey(this.clientEnv.appKey);
                this.client.setDisplayName(this.accountSettings.displayName);
                this.client.login(this.accountSettings.userId, '123456', loginParam);
            }
        },

        sendOnlineMessage() {
            this.client.sendOnlineMessage(this.onlineMessage.content, this.onlineMessage.userId);
        },

        mediaDeviceInit() {
            this.mediaDevice = JRTCMediaDevice.create({
                onVolumeChanged: (volume, userId)=> {
                    if (userId == null || userId == this.client.getUserId()) {
                        let progress = this.volumeMap['progress'+this.client.getUserId()];
                        if (progress) {
                            progress.value = volume / 100;
                        }
                    } else {
                        let progress = this.volumeMap['progress'+userId];
                        if (progress) {
                            progress.value = volume / 100;
                        }
                    }
                },
                onCameraUpdate: ()=> {

                }
            }, {});
        },

        initGuestCallbacks() {
            this.guestCallback = {
                onCallStateChanged: (type, incomingType, inviter, reason)=> {
                    this.onCallStatusChanged(type);
                    this.isTalking = type == GuestCallStateChangeType.TALKING;
                    console.log('sample onCallStateChanged type: ', type, incomingType, inviter, reason);
                    if (type == GuestCallStateChangeType.INCOMING) {
                        let displayName = inviter.userId;
                        if (inviter && inviter.displayName) {
                            displayName = inviter.displayName;
                        }
                        let message =  '收到' + displayName + '来电';
                        if (inviter.extraInfo) {
                            message += '\n' + 'extraInfo：'+ inviter.extraInfo;
                        }
                        result = confirm(message);

                        if (result) {
                            this.guest.answer();
                        } else {
                            this.guest.term();
                        }
                    } else if (type == GuestCallStateChangeType.TALKING) {
                        this.queueInfo = '';
                        Promise.all(
                            [this.startCameraVideo(),
                            this.startLocalAudio()]
                        ).then(() => {
                            this.publishingVideo = true;
                            this.publishingAudio = true;
                            this.getDevices();
                        });
                        this.getParticipants(); // 获取成员

                        if (this.callAgentSettings.isMultiStream) {
                            for (const p of this.roomSettings.participants) {
                                if (p.userId !== this.client.getUserId()) {
                                    this.subscribe(p);//订阅已经在通话内的成员视频
                                }
                            }
                        }

                    } else if (type == GuestCallStateChangeType.TERMED) {
                        this.queueInfo = '';
                        this.publishingVideo = false;
                        this.publishingAudio = false;
                        this.volumeMap = {};
                        this.isUrgent = false;

                        clearInterval(this.statsInternalKey);
                        this.statsInfo = '';

                        this.clearStreamNode();
                    }
                },

                onGetAllGroups: (result, groups)=> {
                    console.log('onGetAllGroups result: ', result, groups);
                    if (groups && groups.length > 0) {
                        this.showToast(JSON.stringify(groups));
                    }
                },

                onCallQueueCount: (count, time, agentRinging)=> {
                    console.log('onCallQueueCount count: %s, time: %s', count, time);
                    this.queueInfo = ' 排队人数: ' + count + ' 等待时长: ' + time;
                    if (this.callState == 'calling') {
                        this.callState = 'waiting'
                    }
                },

                onUrgentResultResponse: (agree) => {
                    console.log('onUrgentResultResponse agree: ', agree);
                    if (!this.isUrgent) {
                        this.isUrgent = agree;
                    }
                },

                onCallPropertyChanged: (propChangeParam)=> {
                    console.log('onCallPropertyChanged propChangeParam: ', propChangeParam);
                    if (propChangeParam.screenShare) {
                        if (this.guest.getShareUserId() && this.guest.getShareUserId().length > 0) {
                            // 屏幕共享渲染
                            this.guest.requestScreenVideo({width: 360, height: 640}).then((result) => {
                                console.debug('requestScreenVideo result', result);
                                let userId = 'screenShare';
                                if (result) {
                                    this.appendRemoteVideo(userId, true);
                                    let viewId = 'video-' + userId;
                                    let box = document.getElementById(`${viewId}`);
                                    const canvas = this.mediaDevice.startVideo(result.streamUrl, this.renderType, result.mediaStream, box);
                                    this.screenShareCanvas = canvas;
                                    if (!canvas) {
                                        this.showToast('屏幕共享视频渲染失败');
                                    }
                                }
                            });
                        } else {
                            this.removeRemoteVideo('screenShare');
                            this.mediaDevice.stopVideo(this.screenShareCanvas);
                            this.screenShareCanvas = undefined;
                        }

                    }
                },

                onMemberJoin: (participant)=> {
                    console.log('onMemberJoin participant: ', participant);
                    if (this.callAgentSettings.isMultiStream) {
                        this.showToast('成员：' + participant.userId + '加入');
                        let exist = false;
                        for (const p of this.roomSettings.participants) {
                            if (p.userId === participant.userId) {
                                exist = true;
                                break;
                            }
                        }
                        if (!exist) {
                            this.subscribe(participant);
                        }
                    } else {
                        if (participant.userId == 'MERGED__') {
                            this.subscribe(participant);
                        } else {
                            this.showToast('成员：' + participant.userId + '加入');
                        }
                    }
                    this.getParticipants();
                },

                onMemberLeave: (participant)=> {
                    this.showToast('成员：' + participant.userId + '离开');
                    this.getParticipants();
                    this.removeRemoteVideo(participant.userId);
                    this.mediaDevice.stopAudioOutput(participant.userId);
                    this.volumeMap['progress' + participant.userId] = null;
                },

                onMemberUpdate: (participant, changeParam)=> {
                    console.log('onMemberUpdate participant: ', participant, changeParam);
                    if (changeParam.netStatus) {
                        let netStatus = participant.netStatus;
                        if (netStatus == NetStatus.VERY_GOOD) {
                            // 网络非常好
                        } else if (netStatus == NetStatus.GOOD) {
                            // 网络好
                        } else if (netStatus == NetStatus.BAD) {
                            // 网络一般
                        } else if (netStatus == NetStatus.NORMAL) {
                            // 网络较差
                        } else if (netStatus == NetStatus.DISCONNECTED) {
                            // 断开连接
                        }
                    }
                },

                onMessageReceived: (content, contentType, messageType, fromUserId)=> {
                    if (contentType == 'InviteScreenShare') {
                        let desc = 'H5无法进行协作';
                        let descMap = {reason_code: 0, reason_desc: desc};
                        this.guest.sendMessage('Decline', JSON.stringify(descMap), fromUserId);
                        return;
                    }
                    let message = "收到消息:" + content + " contentType:" + contentType + " fromUserId:" + fromUserId;
                    this.showToast(message);
                },
                onNotifyMessageReceived: (notifyMessage, fromUserId)=> {
                    console.log('onNotifyMessageReceived notifyMessage: ', notifyMessage, fromUserId);
                },
                onSignRequest: (userId, extraInfo) => {
                    let message = "收到 " + userId + "签名请求 " + extraInfo;
                    this.showToast(message);

                    setTimeout(() => {
                        this.guest.responseSignResult(false, userId, 'H5不支持签名');
                    }, 2000);
                },
                onHoldStateChanged: (hold) => {
                    let message = hold ? "通话被保持" : "通话被取回"
                    this.showToast(message);
                },
                onCallTypeChanged: (callType) => {
                    let message = callType === CallType.AUDIO ? "收到语音通话" : "收到视频通话"
                    this.showToast(message);
                },
                onOnewayVideoChanged: (turnOn) => {
                    let message = turnOn ? "收到打开单项视频" : "收到关闭单项视频"
                    this.showToast(message);
                },
            }
        },

        guestInit() {

            if (this.guestCallback == null) {
                this.initGuestCallbacks();
            }

            let initParam = new JRTCCallCenterInitParam();
            initParam.roomServer = this.roomEnv.server;
            initParam.protocol = 'WEBRTC';

            console.log('tinitParam:',initParam);

            this.guest = JRTCGuest.create(this.client, this.mediaDevice, initParam, this.guestCallback);
        },

        /* page actions  */ 

        queryGroups() {
            if (this.guest == null) {
                this.guestInit();
           }
            this.guest.queryAllGroups();
        },

        startCameraVideo() {

            return new Promise((resolve, reject) => {
                console.log("startCameraVideo");
        
                let viewId = 'local-video';
                this.appendLocalVideo(viewId);
            
                // this.mediaDevice.startCameraVideo(RenderType.RENDER_FULL_SCREEN).then((canvas) => {
                //     let box = document.querySelector(`#${viewId}`);
                //     console.log('box:', box);
                //     if (!this.localCanvas || this.localCanvas.id != canvas.id) {
                //         if (this.localCanvas) {
                //             box?.removeChild(this.localCanvas.videoDivDOM);
                //         }
                //         this.localCanvas = canvas;
                //         box?.appendChild(canvas.videoDivDOM);
                //     }
                // });

                // 或者传入viewId和document
                let box = document.querySelector(`#${viewId}`);
                this.mediaDevice.startCameraVideo(this.renderType, box).then((canvas) => {
                    if (!canvas) {
                        this.showToast('本地视频渲染失败');
                    }
                    this.localCanvas = canvas;
                    resolve();
                });
            })
        },

        stopCameraVideo() {
            console.log("stopCameraVideo", this.localCanvas);
            if (this.localCanvas != null) {
                this.mediaDevice.stopVideo(this.localCanvas);
            }
        },

        startLocalAudio() {
           return new Promise((resolve, reject) => {
                this.mediaDevice.startAudioInput().then((_) => {
                    resolve();
                });
            })
        },

        // 处理摄像头切换
        handleCameraChange() {
            let camera = this.roomSettings.currentCamera;
            this.mediaDevice.switchCamera(camera);
        },

        handleCall(callType) {

            JRTCTracking.setBusinessId("请设置businessId");

            // 设置媒体参数
            if (this.mediaSettings.isSelfDefined) {
                this.callAgentSettings.videoDefinition = 0;
                if (this.mediaSettings.captureResolution == 0) {
                    this.callAgentSettings.captureWidth = 360;
                    this.callAgentSettings.captureHeight = 640;
                } else if (this.mediaSettings.captureResolution == 1) {
                    this.callAgentSettings.captureWidth = 480;
                    this.callAgentSettings.captureHeight = 720;
                } else if (this.mediaSettings.captureResolution == 2) {
                    this.callAgentSettings.captureWidth = 720;
                    this.callAgentSettings.captureHeight = 1080;
                }
                if (!this.callAgentSettings.frameRate | this.callAgentSettings.frameRate <= 0) {
                    this.callAgentSettings.frameRate = 24;
                }
                this.callAgentSettings.svcResolution = this.mediaSettings.svcResolution;
            } else {
                if (this.mediaSettings.maxResolution == 0) {
                    this.callAgentSettings.captureWidth = 360;
                    this.callAgentSettings.captureHeight = 640;
                    this.callAgentSettings.videoDefinition = VideoDefinition.FLUENCY_FRAME_HIGH;
                } else if (this.mediaSettings.maxResolution == 1) {
                    this.callAgentSettings.captureWidth = 480;
                    this.callAgentSettings.captureHeight = 720;
                    this.callAgentSettings.videoDefinition = VideoDefinition.STANDARD_FRAME_HIGH;
                } else if (this.mediaSettings.maxResolution == 2) {
                    this.callAgentSettings.captureWidth = 720;
                    this.callAgentSettings.captureHeight = 1080;
                    this.callAgentSettings.videoDefinition = VideoDefinition.HIGH_FRAME_HIGH;
                }
                this.callAgentSettings.frameRate = 24;
                this.callAgentSettings.maxResolution = this.mediaSettings.maxResolution;
            }

            let callParam = new JRTCCallCenterCallParam();
            callParam.autoRecord = this.callAgentSettings.autoRecord;
            callParam.svcResolution = this.callAgentSettings.svcResolution;
            callParam.maxResolution = this.callAgentSettings.maxResolution;
            callParam.maxFrameRate = this.callAgentSettings.frameRate;
            callParam.securityType = this.callAgentSettings.securityType;
            callParam.extraInfo = this.callAgentSettings.extraInfo;
            callParam.multiStream = this.callAgentSettings.isMultiStream;

            callParam.videoEncodeType = this.callAgentSettings.videoEncodeType;
            callParam.audioEncodeType = this.callAgentSettings.audioEncodeType;
            callParam.heartbeatTime = this.callAgentSettings.heartbeatTime;
            callParam.heartbeatTimeout = this.callAgentSettings.heartbeatTimeout;

            callParam.videoDefinition = this.callAgentSettings.videoDefinition;
            callParam.vipPower = this.callAgentSettings.vipPower;
            callParam.traceId = this.callAgentSettings.traceId;

             // 设置摄像头采集
             this.mediaDevice.setCameraProperty(this.callAgentSettings.captureWidth, this.callAgentSettings.captureHeight, this.callAgentSettings.frameRate);

            if (!this.guest) {
                 this.guestInit();
            }

            let telNumberOrAgentUserId;

            if (callType == 'oneToOneCall') {
                telNumberOrAgentUserId = this.callAgentSettings.agentUserId;
                this.guest.oneToOneCall(telNumberOrAgentUserId, callParam);
            } else if (callType == 'call') {
                telNumberOrAgentUserId = this.callAgentSettings.telNumber;
                this.guest.call(telNumberOrAgentUserId, callParam);
            }
        },

        requestUrgent() {
            this.guest.requestUrgent();
        },
        
        term() {
            if (this.guest != null) {
                this.guest.term();
            }
        },

        getParticipants() {
            const parts = this.guest.getParticipants();
            this.roomSettings.participants = [];
            for (let index = 0; index < parts.length; index++) {
                const element = parts[index];
                this.roomSettings.participants.push(element);
            }
        },

        subscribe(participant) {
            this.guest.requestVideo(participant, {width: 360, height: 640}).then((result) => {
                console.debug('requestVideo result', result);
                let userId = participant.userId;
                
                if (result) {
                    this.appendRemoteVideo(userId, false);
                    let viewId = 'video-' + userId;
                    let box = document.getElementById(`${viewId}`);
                    const canvas = this.mediaDevice.startVideo(participant.streamUrl, this.renderType, result.mediaStream, box);
                    if (!canvas) {
                        if (this.callAgentSettings.isMultiStream) {
                            this.showToast(`${userId} 视频渲染失败`);
                        } else {
                            this.showToast("视频渲染失败");
                        }
                    }
                    this.mediaDevice.startAudioOutput(result.mediaStream, userId);
                } else {
                    if (this.callAgentSettings.isMultiStream) {
                        this.showToast(`${userId} 视频流请求失败`);
                    } else {
                        this.showToast("视频流请求失败");
                    }
                }
            });
        },

        publishVideo() {
            if (this.publishingVideo) {
                return;
            }
            this.publishingVideo = this.guest.enableUploadVideoStream(true);
        },

        unPublishVideo() {
            if (!this.publishingVideo) {
                return;
            }
            this.publishingVideo = false;
            console.log('sample unPublish enableUploadVideoStream');
            this.guest.enableUploadVideoStream(false);
        },

        publishAudio() {
            if (this.publishingAudio) {
                return;
            }
            this.publishingAudio = this.guest.enableUploadAudioStream(true);
        },

        unPublishAudio() {
            if (!this.publishingAudio) {
                return;
            }
            this.guest.enableUploadAudioStream(false);
            this.publishingAudio = false;
        },

        startRemoteRecord() {
            this.guest.enableRemoteRecord(true, this.roomSettings.remoteRecordSettings);
        },

        stopRemoteRecord() {
            this.guest.enableRemoteRecord(false, {});
        },

        sendMessage() {
            // 消息内容
            if (this.roomSettings.sendMessage.content) {
                let size = this.strSize(this.roomSettings.sendMessage.content);
                if (size > 4096) {
                    this.showToast('消息内容过长，不能大于4KB');
                    return; 
                }
            }

            if (this.roomSettings.sendMessage.userId == '' || !this.roomSettings.sendMessage.userId) {
                this.guest.sendMessage(this.roomSettings.sendMessage.type, this.roomSettings.sendMessage.content, null);
            } else {
                this.guest.sendMessage(this.roomSettings.sendMessage.type, this.roomSettings.sendMessage.content, this.roomSettings.sendMessage.userId);
            }
        },

        getDevices() {
            setTimeout(() => {
                this.cameraList = this.mediaDevice.getCameras();

                console.log('this.cameraList = ', this.cameraList);

                if (this.cameraList.length > 0) {
                    this.roomSettings.currentCamera = this.cameraList[0];
                    console.log('this.currentCamera = ', this.roomSettings.currentCamera);
                }

                this.microphoneList = this.mediaDevice.getAudioInputs();

                console.log("this.microphoneList = ", this.microphoneList);

                if (this.microphoneList.length > 0) {
                    this.roomSettings.currentMic = this.microphoneList[0];
                    console.log('this.currentMic = ', this.roomSettings.currentMic);
                }
            }, 500);
        },

        /**
         * helper function
         */
         showToast(msg, duration) {
            this.toastList.push(msg);

            let length = this.toastList.length;
            let top = 5 * length  + 30;

            duration = isNaN(duration) ? 3000 : duration;
            var m = document.createElement('div');
            m.innerHTML = msg;
            m.style.cssText = `width:50%; min-width:180px; background:#000; opacity:0.5; height:auto;min-height: 30px; color:#fff; line-height:30px; text-align:center; border-radius:4px; position:fixed; top:${top}%; left:20%; z-index:999999;`
            document.body.appendChild(m);

            let that = this;
            setTimeout(function() {
                var d = 0.5;
                m.style.webkitTransition = '-webkit-transform ' + d + 's ease-in, opacity ' + d + 's ease-in';
                m.style.opacity = '0';

                setTimeout(function() {
                        document.body.removeChild(m);
                        that.toastList.splice(0, 1);
                    }, d * 1000);
            }, duration);
        },

        onClientStatusChanged(state) {
            if (state == ClientState.LOGINING) {
                this.clientStateText = '登录中';
            } else if (state == ClientState.LOGINED) {
                this.clientStateText = '已登录';
            } else if (state == ClientState.LOGOUTING) {
                this.clientStateText = '登出中';
            } else {
                this.clientStateText = '未登录';
            }
        },

        onCallStatusChanged(type) {
            if (type == GuestCallStateChangeType.CALLING) {
                this.callState = 'calling';
                this.callStateText = '正在呼叫...';
            } else if (type == GuestCallStateChangeType.WAITING) {
                this.callState = 'waiting';
                this.callStateText = '呼叫等待中...';
            } else if (type == GuestCallStateChangeType.INCOMING) {
                this.callState = 'incoming';
                this.callStateText = '收到来电';
            } else if (type == GuestCallStateChangeType.TALKING) {
                this.callState = 'talking';
                this.callStateText = '通话中';
            } else if (type == GuestCallStateChangeType.TERMED) {
                this.callState = 'termed';
                this.callStateText = '通话结束';
            }
        },

        isScreenShareParticipant(participant) {
            if (participant.userId === this.guest.getCallId()) {
             return true; 
            }
            return false;
        },

        clearStreamNode() {
            const remote = document.getElementById("remote-video-container");
            if (remote != null) {
              this.clearAllNode(remote);
            }
            
            const local = document.getElementById("local-video-container");
            if (local != null) {
              this.clearAllNode(local);
            }
        },
    
        clearAllNode(parentNode){
            parentNode.innerHTML = '';
        },

        appendErrorContent(content) {
            let div =  this.createDivDOM();
            div.textContent = content;
            let viewId = 'error-code-detail'
            const box = document.querySelector(`#${viewId}`);
            box.appendChild(div);
        },

        createDivDOM() {
            const divDom = document.createElement('div');
            divDom.style.display = 'flex';
            divDom.style.margin = '0 auto';
            divDom.style.fontSize = '10px';
            return divDom;
        },

        appendLocalVideo(vid) {

            let userId = this.client.getUserId();

            let viewId = 'local-video-container'
            let local_video_container = document.querySelector(`#${viewId}`);
            const figure = document.createElement('figure');

            let div = this.createDivDOM();
            div.style.width = '180px';
            div.style.height = '320px';
            div.id = vid;
            div.style.backgroundColor = 'rgba(0,0,0)';
            figure.appendChild(div);

            let textDiv = this.createDivDOM();
            textDiv.style.width = '180px';
            textDiv.textContent = 'self';
            textDiv.style.display = 'flex';
            textDiv.style.justifyContent = 'center';
            figure.appendChild(textDiv);

            local_video_container.appendChild(figure);
        },

        appendRemoteVideo(userId, isScreenShare) {

            let viewId = 'remote-video-container'
            let remote_video_container = document.querySelector(`#${viewId}`);
            const figure = document.createElement('figure');
            
            figure.className = 'stream-wrapper';
            figure.id = 'figure-'+userId;

            let div = this.createDivDOM();
            div.style.width = '180px';
            div.style.height = '320px';
            div.id = 'video-' + userId;
            div.style.backgroundColor = 'rgba(0,0,0)';
            figure.appendChild(div);

            if (!isScreenShare) {
                const progress = document.createElement('progress');
                progress.style.display = 'flex';
                progress.style.margin = '0 auto';
                this.volumeMap['progress'+userId] = progress;
                figure.appendChild(progress);
            }

            let textDiv = this.createDivDOM();
            textDiv.style.width = '180px';
            textDiv.textContent = userId;
            textDiv.style.display = 'flex';
            textDiv.style.justifyContent = 'center';
            figure.appendChild(textDiv);

            remote_video_container.appendChild(figure);
        },

        removeRemoteVideo(userId) {
            let viewId = 'remote-video-container';
            let remote_video_container = document.querySelector(`#${viewId}`);

            let videoViewId = 'figure-' + userId;
            let video = document.getElementById(`${videoViewId}`);
            if (video) {
                remote_video_container.removeChild(video);
            }
        },
        openFAQPage() {
            window.open(`./FAQ.html`);
        },
        // 计算字符串占用内存
        strSize(str, charset) {
            let total = 0;
            charset = charset?.toLowerCase() || '';
            for (i = 0; i < str.length; i++) {
                let charCode = str.charCodeAt(i);
                if (charset === 'utf-16' || charset === 'utf16') {
                    total += charCode <= 0xffff ? 2 : 4;
                } else {
                    if (charCode <= 0x007f) {
                        total += 1;
                    } else if (charCode <= 0x07ff) {
                        total += 2;
                    } else if (charCode <= 0xffff) {
                        total += 3;
                    } else {
                        total += 4;
                    }
                }
            }
            return total;
        }
    
    }
  }).mount('#app');