
const {
    JRTCClient, 
    JRTCClientInitParam, 
    JRTCClientLoginParam,
    JRTCClientCallback,
    ClientState,
    JRTCMediaDevice,
    RenderType,
    JRTCMediaDeviceVideoCanvas,
    JRTCRoom, 
    JRTCRoomInitParam, 
    JRTCRoomJoinParam,
    JRTCRecordRemoteParam,
    JRTCMergeVideoParam,
    StompSocketAuthType,
    RoomState,
    JRTCRoomParticipant,
    RemoteRecordState,
    JRTCVersion,
    JRTCTracking,
    StorageManager,
    IntelligentMergeMode,
    VideoMergeMode,
    ScsMergeMode,
    JRTCLog,
    VideoDefinition,
    NetStatus,
    MediaType
} = JRTCSDK;

function openGuest() {
    window.open("../index.html", '_self');
}

function openAutoTest() {
    window.open("../autotest/index.html", '_self');
}

const MergeUserId = 'MERGED__';

const app = Vue.createApp({

    data() {
      return {

        //client的环境参数
        clientEnv: {
            server: 'https://192.168.4.171:10034',
            appKey: '05c97ce0d0a84d2592334093',
            token: '',
            tokenType: 'jrtc_access'
        },

        //房间的环境参数
        roomEnv: {
            server: 'https://192.168.4.171:10034',
        },

        /**
         * client
         */
        client: null,

        clienstState: ClientState.NOT_INIT,

        // 是否初始化
        isInitialized: false,

        clientStateText: '未登录',

        // 是否登录
        isLogin: false,

        version: '',

        /**
         * mediaDevice
         */
        mediaDevice: null,

        cameraList: [],

        microphoneList: [],
        
        localCanvas : null,

        localAuido: null,

        // 音量 progress map
        volumeMap: {},
        
        renderTypeMap: [
            {'label': 'FullContent', value: RenderType.RENDER_FULL_CONTENT},
            {'label': 'FullScreen', value: RenderType.RENDER_FULL_SCREEN},
        ],

        // 渲染模式
        renderType: RenderType.RENDER_FULL_CONTENT,

        /**
         * room
         */
        room: null,

        roomCallback: null,

        roomStatus: 'idle',

        roomStateText: '空闲',

        roomState: RoomState.IDLE,

        isJoined: false,

        maxResolutionMap: [
            {'label': '360p', value: 0},
            {'label': '720p', value: 1},
            {'label': '1080p', value: 2},
        ],

        securityTypeMap: [
            {'label': '不加密', value: 0},
            {'label': 'SRTP', value: 1},
            {'label': 'SM4', value: 2},
        ],

        videoEncodeTypeMap: [
            {'label': 'H264', value: 0},
            {'label': 'AV1', value: 1},
        ],

        audioEncodeTypeMap: [
            {'label': 'OPUS', value: 0},
            {'label': 'PCMA', value: 1},
            {'label': 'PCMU', value: 2},
        ],

        recordMergeModeMap: [
            {'label': '初始值', value: VideoMergeMode.INIT},
            {'label': '自定义布局', value: VideoMergeMode.CUSTOM_LAYOUT},
            {'label': '智能分屏', value: VideoMergeMode.INTELLIGENT_LAYOUT},
        ],

        intelligentMergeModeMap: [
            {'label': '初始值', value: IntelligentMergeMode.INIT},
            {'label': '自由布局', value: IntelligentMergeMode.FREE_LAYOUT},
            {'label': '矩形布局', value: IntelligentMergeMode.RECT_LAYOUT},
            {'label': '大小屏放大2倍', value: IntelligentMergeMode.BIG_SMALL_X2},
            {'label': '大小屏放大3倍', value: IntelligentMergeMode.BIG_SMALL_X3},
            {'label': '大小屏放大4倍', value: IntelligentMergeMode.BIG_SMALL_X4},
            {'label': '大小屏放大5倍', value: IntelligentMergeMode.BIG_SMALL_X5},
            {'label': '大小屏放大6倍', value: IntelligentMergeMode.BIG_SMALL_X6},
            {'label': '大小屏放大7倍', value: IntelligentMergeMode.BIG_SMALL_X7},
            {'label': '大小屏放大8倍', value: IntelligentMergeMode.BIG_SMALL_X8},
            {'label': '大小屏放大9倍', value: IntelligentMergeMode.BIG_SMALL_X9},
            {'label': '大小屏放大10倍', value: IntelligentMergeMode.BIG_SMALL_X10},
            {'label': '大小屏放大到顶部', value: IntelligentMergeMode.BIG_SMALL_TOP},
            {'label': '大小屏放大到底部', value: IntelligentMergeMode.BIG_SMALL_BOTTOM},
            {'label': '绝对等分模式', value: IntelligentMergeMode.DIVIDE_LAYOUT},
            {'label': '绝对等分模式不带图像角度', value: IntelligentMergeMode.DIVIDE_LAYOUT_NO_ANGLE},
        ],

        scsMergeModeMap: [
            {'label': '初始值', value: ScsMergeMode.INIT},
            {'label': '屏幕共享', value: ScsMergeMode.SCREEN_SHARE},
            {'label': '将屏幕共享和所有视频内容合并，所有视频不覆盖屏幕共享', value: ScsMergeMode.PLATE_FORM},
            {'label': '将屏幕共享和所有视频内容合并，所有视频覆盖在屏幕共享', value: ScsMergeMode.PLATE_FORM_COVER},
            {'label': '将屏幕共享和焦点视频内容合并', value: ScsMergeMode.SPEAKER},
            {'label': '忽略屏幕共享', value: ScsMergeMode.WITHOUT_SCREEN_SHARE},
        ],

        onlineMessage: {
            content: '',
            userId: ''
        },

        mediaSettings: {
            isSelfDefined: false,
            maxResolution: 0,
            svcResolution: '1 90 70 180 180 360 480 720 1350',
            captureResolution: 1,
            frameRate: 24
        },

        joinSettings: {
            userId: '',
            displayName: '',
            roomId: '',
            password: '',
            capacity: 6,
            enableRecord: true,
            isMultiStream: false,
            maxResolution: 0,
            uploadLocalVideo: true,
            uploadLocalAudio: true,
            wholeRatio: 1.7777,
            securityType: 0,
            smoothMode: true,
            heartbeatTime: 20,
            heartbeatTimeout: 60,
            videoEncodeType: 0,
            audioEncodeType: 0,
            extraInfo: '',
            // 
            captureWidth: 360,
            captureHeight: 640,
            frameRate: 24,
            videoDefinition: 0, // VideoDefinition 
            traceId: '' //业务流水号
        },

        roomSettings: {
            currentCamera: null,
            currentMic: null,
            remoteMergeVideoSettings: {
                frameRate: 24,
                bitrate: 0,
                width: 640,
                height: 360,
                intelligentMergeMode: IntelligentMergeMode.INIT, // 智能分屏无屏幕共享
                scsMergeMode: ScsMergeMode.INIT, // 智能分屏带屏幕共享
            },
            remoteRecordSettings: {
                recordVideo: true,
                mergeMode: VideoMergeMode.INIT, // 合并模式
                intelligentMergeMode: IntelligentMergeMode.INIT, // 智能分屏无屏幕共享
                scsMergeMode: ScsMergeMode.INIT, // 智能分屏带屏幕共享

                frameRate: 24,
                bitrate: 15,
                
                videoWidth: 640,
                videoHeight: 360,

                waterMark: '',
                extraInfo: '',
                layoutType: '',
                layoutArray: '',

                fileName: 'H5_record',

                splitFileSize: -1,
            },
            queryRoom: {
                roomId: null,
            },
            sendMessage: {
                type: '',
                content: '',
                userId: '',
            },
            participants: [],

            screenShareUserId: '',
            recordState: RemoteRecordState.NONE,

            enableStopRecord: false,
            ratio: 0.5625,

        },

        publishingVideo: false, // 是否正在视频推流
        publishingAudio: false, // 是否正在音频推流

        screenShareCanvas: null,

        /**
         * Stats
         */
        statsInfo: '',
        statsInternalKey: null,

        /**
         * storage
         */

        storage: null,

        /**
         * toast
         */

        toastList: [],

        // 日志存储
        logStore: true

      };
    },
    created() {

        this.version = JRTCVersion.getSDKVersion();

        this.storage = new StorageManager();

        const settingFormData = this.storage.get('settingFormData_webrtc-room-sdk');

        if (settingFormData) {
            Object.entries(settingFormData)
                .forEach((entity) => this[entity[0]] = entity[1]);
        }

        window.addEventListener('beforeunload', (ev) => {
            const {roomSettings, joinSettings, mediaSettings, clientEnv, roomEnv, logStore} = this;

            this.storage.set('settingFormData_webrtc-room-sdk', {
                roomSettings,
                joinSettings,
                mediaSettings,
                clientEnv,
                roomEnv,
                logStore
            });

            ev.preventDefault();
            ev.returnValue = '';
            return '';
        });

        window.addEventListener('pagehide', (ev) => {
            const {roomSettings, joinSettings, mediaSettings, clientEnv, roomEnv, logStore} = this;

            this.storage.set('settingFormData_webrtc-room-sdk', {
                roomSettings,
                joinSettings,
                mediaSettings,
                clientEnv,
                roomEnv,
                logStore
            });

            ev.preventDefault();
            ev.returnValue = '';
            return '';
        });

        if (!this.joinSettings.captureWidth | this.joinSettings.captureWidth <= 0) {
            this.joinSettings.captureWidth = 360;
        }

        if (!this.joinSettings.captureHeight | this.joinSettings.captureHeight <= 0) {
            this.joinSettings.captureWidth = 640;
        }

        if (!this.joinSettings.frameRate | this.joinSettings.frameRate <= 0) {
            this.joinSettings.frameRate = 24;
        }
        if (!this.roomSettings.remoteMergeVideoSettings) {
            this.roomSettings.remoteMergeVideoSettings = {
                frameRate: 24,
                bitrate: 0,
                width: 640,
                height: 360,
                intelligentMergeMode: IntelligentMergeMode.INIT, // 智能分屏无屏幕共享
                scsMergeMode: ScsMergeMode.INIT, // 智能分屏带屏幕共享
            }
        }
    },

    methods: {

        trust() {
            window.open(this.clientEnv.server, '_blank');
        },

        /**
         * client
         */
        clientInit() {
            const clientInitParam = new JRTCClientInitParam();
            clientInitParam.appName = "JCCSample-H5";
            clientInitParam.server = this.clientEnv.server;
            clientInitParam.appKey = this.clientEnv.appKey;
            this.clienstState = ClientState.IDLE;

            this.client = JRTCClient.create({
                onLogin: (result, reason) => {
                    console.log('onLogin result: ', result, ' reason: ', reason);
                },
                onLogout: (reason) => {
                    console.log('onLogout reason: ', reason);
                    this.leave();
                },
                onClientStateChanged: (state, oldState) => {
                    console.log('onClientStateChanged state: ', state);
                    this.clienstState = state;
                    this.isLogin = state == ClientState.LOGINED || state == ClientState.LOGINING;

                    this.onClientStatusChanged(state);
                },
                onSDKEvent: (event) => {
                    console.log('onSDKEvent event: ', event);
                    if (event.errorCode && event.errorCode.length > 0 && event.errorCodeDetail && event.errorCodeDetail.length > 0) {
                        let content = "错误码:" + event.errorCode + "错误描述:" + event.errorCodeDetail;
                        this.appendErrorContent(content);
                    }
                },
                onOnlineMessageReceived: (message, userId) => {
                    this.showToast('收到来自 ' + userId + ' 发送的在线消息: ' + message);
                },
                onOnlineMessageSendResult: (result, operatorId) => {
                    let content = '发送';
                    content += result ? '成功' : '失败';
                    content += ' operatorId: ' + operatorId;
                    console.log(content);
                }
            }, clientInitParam);
            JRTCTracking.setChannel("H5 channel");

            JRTCLog.localLogStore = this.logStore;
        },

        initHandle() {
            if (this.isInitialized) {
                if (this.isLogin) {
                    this.client.logout();
                }
                JRTCClient.destroy();
                this.client = null;

                JRTCMediaDevice.destroy();
                this.mediaDevice = null;

                if (this.isJoined) {
                    this.room.leave();
                }
                JRTCRoom.destroy();
                this.room = null;
            } else {
                this.clientInit();
                this.mediaDeviceInit();
            }
            this.isInitialized = !this.isInitialized;
        },

        downloadLog() {
            
            const blob = JRTCLog.getLogMessage(),
            blobUrl = window.URL.createObjectURL(blob),
            tempLink = document.createElement('a');

            tempLink.style.display = 'none';
            tempLink.href = blobUrl;
            tempLink.setAttribute('download', `H5 ${window.location.host}-${Date.now()}.log`);

            if (typeof tempLink.download === 'undefined') {
                tempLink.setAttribute('target', '_blank');
            }

            document.body.appendChild(tempLink);
            tempLink.click();
            document.body.removeChild(tempLink);

            setTimeout(() => {
                window.URL.revokeObjectURL(blobUrl);
            });
        },

        clearLog() {
            JRTCLog.clearLogMessage();
        },

        loginHandle() {
            if (this.isLogin) {
                this.client.logout();
            } else {
                this.clientInit();
                let loginParam = new JRTCClientLoginParam();
                loginParam.token = this.clientEnv.token;
                loginParam.tokenType = this.clientEnv.tokenType;
                this.client.setDisplayName(this.joinSettings.displayName);
                this.client.login(this.joinSettings.userId, '123456', loginParam);
            }
        },

        sendOnlineMessage() {
            this.client.sendOnlineMessage(this.onlineMessage.content, this.onlineMessage.userId);
        },

        mediaDeviceInit() {
            this.mediaDevice = JRTCMediaDevice.create({
                onVolumeChanged: (volume, userId)=> {
                    if (userId == null || userId == this.client.getUserId()) {
                        let progress = this.volumeMap['progress'+this.client.getUserId()];
                        if (progress) {
                            progress.value = volume / 100;
                        }
                    } else {
                        let progress = this.volumeMap['progress'+userId];
                        if (progress) {
                            progress.value = volume / 100;
                        }
                    }
                },
                onCameraUpdate: ()=> {

                },
                onAudioInputUpdate: ()=> {

                }
            }, {});

            setTimeout(() => {
                this.cameraList = this.mediaDevice.getCameras();

                console.debug('this.cameraList = ', this.cameraList);

                if (this.cameraList.length > 0) {
                    this.roomSettings.currentCamera = this.cameraList[0];
                    console.debug('this.currentCamera = ', this.roomSettings.currentCamera);
                }

                this.microphoneList = this.mediaDevice.getAudioInputs();

                console.log('this.microphoneList:', this.microphoneList);

                if (this.microphoneList.length > 0) {
                    this.roomSettings.currentMic = this.microphoneList[0];

                    console.log('this.roomSettings.currentMic:', this.roomSettings.currentMic);
                }
            }, 500);
        },

        /* page actions  */ 
        startCamera() {
            console.log("startCamera");
            this.mediaDevice.startCamera();
        },

        stopCamera() {
            console.log("stopCamera");
            this.mediaDevice.stopCamera();
        },

        startCameraVideo() {
            return new Promise((resolve, reject) => {
                console.log("startCameraVideo");
        
                let viewId = 'local-video';
                this.appendLocalVideo(viewId);
            
                // this.mediaDevice.startCameraVideo(RenderType.RENDER_FULL_SCREEN).then((canvas) => {
                //     let box = document.querySelector(`#${viewId}`);
                //     console.log('box:', box);
                //     if (!this.localCanvas || this.localCanvas.id != canvas.id) {
                //         if (this.localCanvas) {
                //             box?.removeChild(this.localCanvas.videoDivDOM);
                //         }
                //         this.localCanvas = canvas;
                //         box?.appendChild(canvas.videoDivDOM);
                //     }
                // });
            
                // 或者传入viewId和document
                let box = document.querySelector(`#${viewId}`);
                this.mediaDevice.startCameraVideo(this.renderType, box).then((canvas) => {
                    if (!canvas) {
                        this.showToast(`本地视频渲染失败`);
                    }
                    this.localCanvas = canvas;
                    resolve();
                });
            })
        },

        stopCameraVideo() {
            if (this.localCanvas != null) {
                this.mediaDevice.stopVideo(this.localCanvas);
            }
        },

        startLocalAudio() {
           return new Promise((resolve, reject) => {
                this.mediaDevice.startAudioInput().then(() => {
                    resolve();
                });
            })
        },

        // 处理设备切换
        handleCameraChange() {
            let camera = this.roomSettings.currentCamera;
            this.mediaDevice.switchCamera(camera);
        },

        /**
         * Room 相关
         */

        roomInit() {
            if (this.roomCallback == null) {
                this.initRoomCallbacks();
            }
            
            const param =  new JRTCRoomInitParam();
            param.roomServer = this.roomEnv.server;
            param.protocol = 'WEBRTC';

            const paramStr = param.toString();
            console.log('param: ', paramStr);
            this.room = JRTCRoom.create(this.client, this.mediaDevice, param, this.roomCallback);
        },

        discoveryRoom() {
            const roomId = this.roomSettings.queryRoom.roomId;
            console.log('discoveryRoom roomId: ', roomId);
            let appKey = this.clientEnv.appKey;
            var oReq = new XMLHttpRequest();

            let that = this;

            oReq.onload = function(e) {
                let response = JSON.parse(oReq.response)
                console.log('oReq.response 11', response);
                let data = response.data;
                if (data != null) {
                    if (data.exist != null && data.exist) {
                        that.showToast(`房间号：${roomId}已经存在`);
                    } else {
                        that.showToast(`房间号：${roomId}不存在`);
                    }
                } else {
                    that.showToast("查询房间失败");
                }
            }
            oReq.open("GET", `${this.roomEnv.server}/room/discovery?roomId=${roomId}&appKey=${appKey}`);
            oReq.send();
        },

        joinRoom() {
            if (!this.isLogin) {
                this.showToast('请先登录');
                return;
            }

            if (this.isJoined) {
                this.showToast('已经加入房间');
                return;
            }
            if (this.room == null) {
                this.roomInit();
            }
    

            // 设置媒体参数
            if (this.mediaSettings.isSelfDefined) {
                this.joinSettings.videoDefinition = 0;
                if (this.mediaSettings.captureResolution == 0) {
                    this.joinSettings.captureWidth = 360;
                    this.joinSettings.captureHeight = 640;
                } else if (this.mediaSettings.captureResolution == 1) {
                    this.joinSettings.captureWidth = 480;
                    this.joinSettings.captureHeight = 720;
                } else if (this.mediaSettings.captureResolution == 2) {
                    this.joinSettings.captureWidth = 720;
                    this.joinSettings.captureHeight = 1080;
                }
                if (!this.joinSettings.frameRate | this.joinSettings.frameRate <= 0) {
                    this.joinSettings.frameRate = 24;
                }
                this.joinSettings.svcResolution = this.mediaSettings.svcResolution;
            } else {
                if (this.mediaSettings.maxResolution == 0) {
                    this.joinSettings.captureWidth = 360;
                    this.joinSettings.captureHeight = 640;
                    this.joinSettings.videoDefinition = VideoDefinition.FLUENCY_FRAME_HIGH;
                } else if (this.mediaSettings.maxResolution == 1) {
                    this.joinSettings.captureWidth = 480;
                    this.joinSettings.captureHeight = 720;
                    this.joinSettings.videoDefinition = VideoDefinition.STANDARD_FRAME_HIGH;
                } else if (this.mediaSettings.maxResolution == 2) {
                    this.joinSettings.captureWidth = 720;
                    this.joinSettings.captureHeight = 1080;
                    this.joinSettings.videoDefinition = VideoDefinition.HIGH_FRAME_HIGH;
                }
                this.joinSettings.frameRate = 24;
                this.joinSettings.maxResolution = this.mediaSettings.maxResolution;
            }

            // 设置摄像头采集
            this.mediaDevice.setCameraProperty(this.joinSettings.captureWidth, this.joinSettings.captureHeight, this.joinSettings.frameRate);
            // 设置媒体类型，默认 ALL
            // this.mediaDevice.setSupportMediaType(MediaType.ALL);

            const param = new JRTCRoomJoinParam();
            param.enableRemoteRecord = true;
            param.capacity = this.joinSettings.capacity;
            param.password = this.joinSettings.password;
            param.svcResolution = this.joinSettings.svcResolution;
            param.maxResolution = this.joinSettings.maxResolution;
            param.maxFrameRate = this.joinSettings.frameRate;
            param.wholeRatio = this.joinSettings.wholeRatio;
            param.securityType = this.joinSettings.securityType;
            param.smoothMode = this.joinSettings.smoothMode;
            param.heartbeatTime = this.joinSettings.heartbeatTime;
            param.heartbeatTimeout = this.joinSettings.heartbeatTimeout;
            param.extraInfo = this.joinSettings.extraInfo;
            param.videoEncodeType = this.joinSettings.videoEncodeType;
            param.audioEncodeType = this.joinSettings.audioEncodeType;

            param.multiStream = this.joinSettings.isMultiStream;

            param.uploadLocalVideo = this.joinSettings.uploadLocalVideo;
            param.uploadLocalAudio = this.joinSettings.uploadLocalAudio;

            param.videoDefinition = this.joinSettings.videoDefinition;
            param.traceId = this.joinSettings.traceId;
            
            const paramStr = param.toString();
            console.log('param: ', paramStr);
            console.log('param.video: ', param.video);
            this.publishingAudio = param.uploadLocalAudio;
            this.publishingVideo = param.uploadLocalVideo;
            this.room.join(this.joinSettings.roomId, param);
        },

        leave() {
            if (this.roomState == RoomState.JOINED || this.roomState == RoomState.JOINING) {
                this.room.leave();
            }
        },

        getParticipants() {
            const parts = this.room.getParticipants();
            console.log('parts: ', parts);
            this.roomSettings.participants = [];
            for (let index = 0; index < parts.length; index++) {
                const element = parts[index];
                this.roomSettings.participants.push(element);
            }
        },
        
        subscribe(participant) {
            this.room.requestVideo(participant, {width: 360, height: 640}).then((result) => {
                console.debug('requestVideo result', result);
                let userId = participant.userId;
                if (result) {
                    let viewId = 'video-' + userId;
                    this.appendRemoteVideo(userId, false);
                    let box = document.getElementById(`${viewId}`);
                    const canvas = this.mediaDevice.startVideo(participant.streamUrl, this.renderType, result.mediaStream, box);
                    if (!canvas) {
                        if (this.joinSettings.isMultiStream) {
                            this.showToast(`${userId} 视频渲染失败`);
                        } else {
                            this.showToast("视频渲染失败");
                        }
                    }
                    this.mediaDevice.startAudioOutput(result.mediaStream, userId);
                } else {
                    if (this.joinSettings.isMultiStream) {
                        this.showToast(`${userId} 视频流请求失败`);
                    } else {
                        this.showToast("视频流请求失败");
                    }
                }
            });
        },

        publishVideo() {
            if (this.room.getUploadLocalVideo()) {
                return;
            }
            this.room.enableUploadVideoStream(true);
            this.publishingVideo = this.room.getUploadLocalVideo();
        },

        unPublishVideo() {
            if (!this.room.getUploadLocalVideo()) {
                return;
            }
            this.room.enableUploadVideoStream(false);
            this.publishingVideo = this.room.getUploadLocalVideo();
        },

        publishAudio() {
            if (this.room.getUploadLocalAudio()) {
                return;
            }
            this.room.enableUploadAudioStream(true);
            this.publishingAudio = this.room.getUploadLocalAudio();
        },

        unPublishAudio() {
            if (!this.room.getUploadLocalAudio()) {
                return;
            }
            this.room.enableUploadAudioStream(false);
            this.publishingAudio = this.room.getUploadLocalAudio();
        },

        switchCamera() {
            this.mediaDevice.switchCamera();
        },

        startRemoteRecord() {
            let param = new JRTCRecordRemoteParam();
            param.recordVideo = this.roomSettings.remoteRecordSettings.recordVideo;
            param.mergeMode = this.roomSettings.remoteRecordSettings.mergeMode;
            param.intelligentMergeMode = this.roomSettings.remoteRecordSettings.intelligentMergeMode;
            param.scsMergeMode = this.roomSettings.remoteRecordSettings.scsMergeMode;

            param.iBitrate = this.roomSettings.remoteRecordSettings.bitrate;
            param.frameRate = this.roomSettings.remoteRecordSettings.frameRate;
            
            param.videoWidth = this.roomSettings.remoteRecordSettings.videoWidth;
            param.videoHeight = this.roomSettings.remoteRecordSettings.videoHeight;

            console.log('kkkk startRemoteRecord waterMark:', this.roomSettings.remoteRecordSettings.waterMark);

            if (this.roomSettings.remoteRecordSettings.waterMark && this.roomSettings.remoteRecordSettings.waterMark.length > 0) {
                let waterMarkMap = JSON.parse(this.roomSettings.remoteRecordSettings.waterMark);
                if (waterMarkMap) {
                    param.watermarkTextMap = waterMarkMap;
                }
            }

            console.log('kkkk startRemoteRecord param.watermarkTextMap:', param.watermarkTextMap);
            
            param.extraInfo = this.roomSettings.remoteRecordSettings.extraInfo;

            param.layoutType = this.roomSettings.remoteRecordSettings.layoutType;

            if (this.roomSettings.remoteRecordSettings.layoutArray && this.roomSettings.remoteRecordSettings.layoutArray.length > 0) {
                let layoutArray = JSON.parse(this.roomSettings.remoteRecordSettings.layoutArray);
                if (layoutArray) {
                    param.layoutArray = layoutArray;
                }
            }

            param.fileName = this.roomSettings.remoteRecordSettings.fileName;
            param.splitFileSize = this.roomSettings.remoteRecordSettings.splitFileSize;
            this.room.enableRemoteRecord(true, param);
        },

        stopRemoteRecord() {
            this.room.enableRemoteRecord(false, {});
        },

        setRemoteMergeVideoProperty() {
            let mergeParam = new JRTCMergeVideoParam();
            mergeParam.width = this.roomSettings.remoteMergeVideoSettings.width;
            mergeParam.height = this.roomSettings.remoteMergeVideoSettings.height;
            mergeParam.frameRate = this.roomSettings.remoteMergeVideoSettings.frameRate;
            mergeParam.iBitrate = this.roomSettings.remoteMergeVideoSettings.bitrate;
            mergeParam.intelligentMergeMode = this.roomSettings.remoteMergeVideoSettings.intelligentMergeMode;
            mergeParam.scsMergeMode = this.roomSettings.remoteMergeVideoSettings.scsMergeMode;
            this.room.setRemoteMergeVideoProperty(mergeParam);
        },

        sendMessage() {
            // 消息内容
            if (this.roomSettings.sendMessage.content) {
                let size = this.strSize(this.roomSettings.sendMessage.content);
                if (size > 4096) {
                    this.showToast('消息内容过长，不能大于4KB');
                    return; 
                }
            }
            if (this.roomSettings.sendMessage.userId == '' || !this.roomSettings.sendMessage.userId) {
                this.room.sendMessage(this.roomSettings.sendMessage.type, this.roomSettings.sendMessage.content, null);
            } else {
                this.room.sendMessage(this.roomSettings.sendMessage.type, this.roomSettings.sendMessage.content, this.roomSettings.sendMessage.userId);
            }
        },

        setRatio() {
            if (this.roomSettings.ratio > 0) {
                this.room.setRatio(this.roomSettings.ratio);
            }
        },

        snapshot() {
            this.localCanvas.snapshot()
            .then(blob => {
                document.getElementById('snapshot-preview').src = window.URL.createObjectURL(blob);
            }).catch(e =>{
                console.error(e);
                that.showToast('截图失败');
              });
        },

        initRoomCallbacks() {
            this.roomCallback = {
                onJoin: (result, reason, roomId) =>  {
                    console.debug('onJoin result: ', result, reason, roomId);
                    if (result) {
                        this.publishingAudio = this.room.getUploadLocalAudio();
                        this.publishingVideo = this.room.getUploadLocalVideo();
                        this.getParticipants(); // 获取成员

                        if (this.joinSettings.isMultiStream) {
                            for (const p of this.roomSettings.participants) {
                                if (p.userId !== this.client.getUserId()) {
                                    this.subscribe(p);//订阅已经在房间内的成员视频
                                }
                            }
                        }

                        Promise.all(
                            // audio video
                            [this.startCameraVideo()
                             ,this.startLocalAudio()]

                            // audio only
                            // [this.startLocalAudio()]

                            // video only
                            // [this.startCameraVideo()]
                        ).then(() => {
                            console.log('sample enableUploadVideoStream');
                            this.getDevices();
                        });
                    }
                },
                onQuery: (operationId, result, reason, queryInfo) => {
                    console.debug('onQuery operationId : result: reason: queryInfo: ', operationId, result, reason, queryInfo);
                    this.showToast('房间信息：' + queryInfo.toString());
                },
                onRoomStateChanged: (state, oldState) => {
                    console.debug('onRoomStateChange state:%s oldState:%s', state, oldState);
                    this.roomState = state;
                    this.isJoined = state == RoomState.JOINED;
                    this.onRoomStatusChanged(state);
                },
                onLeave: (reason, roomId) =>  {
                    console.log('onLeave');
                    this.roomSettings.participants = [];
                    this.volumeMap = {};

                    this.roomSettings.recordState = RemoteRecordState.NONE;
                    this.roomSettings.enableRecord = false;
                    this.roomSettings.enableStopRecord = false;
                    this.publishingVideo = false;
                    this.publishingAudio = false;

                    clearInterval(this.statsInternalKey);

                    this.statsInfo = '';

                    this.clearStreamNode();
                },
                onParticipantJoin: (participant) => {
                    console.debug('onParticipantJoin participant: ', participant);
                    if (this.joinSettings.isMultiStream) {
                        this.showToast('成员：' + participant.userId + '加入');
                        let exist = false;
                        for (const p of this.roomSettings.participants) {
                            if (p.userId === participant.userId) {
                                exist = true;
                                break;
                            }
                        }
                        if (!exist) {
                            this.subscribe(participant);
                        }
                    } else {
                        if (participant.userId == MergeUserId) {
                            this.subscribe(participant);
                        } else {
                            if (this.roomSettings.participants.length == 1) {
                                this.showRemoteVideo(true);
                            }
                            this.showToast('成员：' + participant.userId + '加入');
                        }
                    }
                    this.getParticipants();
                },
                onParticipantLeft: (participant) => {
                    this.showToast('成员：' + participant.userId + '离开');
                    this.getParticipants();
                    if (!this.joinSettings.isMultiStream) {
                        if (this.roomSettings.participants.length == 1) {
                            this.showRemoteVideo(false);
                        }
                    } else {
                        this.removeRemoteVideo(participant.userId);
                    }
                    this.mediaDevice.stopAudioOutput(participant.userId);
                    this.volumeMap['progress' + participant.userId] = null;
                },

                onParticipantUpdate: (participant, changeParam) => {
                    console.debug('onParticipantUpdate participant: ', participant, changeParam);
                    if (changeParam.netStatus) {
                        let netStatus = participant.netStatus;
                        if (netStatus == NetStatus.VERY_GOOD) {
                            // 网络非常好
                            console.debug('网络非常好 ');
                        } else if (netStatus == NetStatus.GOOD) {
                            // 网络好
                        } else if (netStatus == NetStatus.BAD) {
                            // 网络一般
                        } else if (netStatus == NetStatus.NORMAL) {
                            // 网络较差
                        } else if (netStatus == NetStatus.DISCONNECTED) {
                            // 断开连接
                        }
                    }
                },
                onRoomPropertyChanged: (param)=> {
                    console.debug('onRoomPropertyChanged param: ', param);
                    if (param.remoteRecordState) {
                        
                        this.roomSettings.recordState = this.room.getRemoteRecordState();
                        this.roomSettings.enableRecord = this.roomSettings.recordState == RemoteRecordState.READY;
                        this.roomSettings.enableStopRecord = this.roomSettings.recordState == RemoteRecordState.RUNNING;
                    }
                    if (param.screenShare) {
                        if (this.room.getShareUserId() && this.room.getShareUserId().length > 0) {
                            // 屏幕共享渲染
                            this.room.requestScreenVideo({width: 360, height: 640}).then((result) => {
                                console.log('requestScreenVideo result', result);
                                let userId = "screenShare";
                                if (result) {
                                    let viewId = 'video-' + userId;
                                    this.appendRemoteVideo(userId, true);
                                    let box = document.getElementById(`${viewId}`);
                                    const canvas = this.mediaDevice.startVideo(result.streamUrl, this.renderType, result.mediaStream, box);
                                    this.screenShareCanvas = canvas;
                                    if (!canvas) {
                                        this.showToast('屏幕共享视频渲染失败');
                                    }
                                } else {
                                    this.showToast(`屏幕共享视频流请求失败`);
                                }
                            });
                        } else {
                            this.removeRemoteVideo('screenShare');
                            this.mediaDevice.stopVideo(this.screenShareCanvas);
                            this.screenShareCanvas = undefined;
                        }
                    }
                },
                onMessageReceived: (type, content, fromUserId) => {
                    if (type == 'JCRemoteDesktop') {
                        let desc = 'H5无法进行协作';
                        let descMap = {reason_code: 0, reason_desc: desc}
                        this.room.sendMessage('Decline', JSON.stringify(descMap), fromUserId);
                        return;
                    }

                    console.debug('onMessageReceived type: ', type, content, fromUserId);
                    let message = "收到消息：" + content + " type：" + type + " fromUserId：" + fromUserId;
                    this.showToast(message);
                }
            }
        },

        isScreenShareParticipant(participant) {
            if (participant.userId === this.room.getCallId()) {
             return true; 
            }
            return false;
        },

        getDevices() {
            setTimeout(() => {
                this.cameraList = this.mediaDevice.getCameras();

                console.debug('this.cameraList = ', this.cameraList);

                if (this.cameraList.length > 0) {
                    this.roomSettings.currentCamera = this.cameraList[0];
                    console.debug('this.currentCamera = ', this.roomSettings.currentCamera);
                }

                this.microphoneList = this.mediaDevice.getAudioInputs();

                console.log('this.microphoneList:', this.microphoneList);

                if (this.microphoneList.length > 0) {
                    this.roomSettings.currentMic = this.microphoneList[0];

                    console.log('this.roomSettings.currentMic:', this.roomSettings.currentMic);
                }
            }, 500);
        },
        /**
         * helper function
         */
        showToast(msg, duration) {
            this.toastList.push(msg);

            let length = this.toastList.length;
            let top = 5 * length  + 30;

            duration = isNaN(duration) ? 3000 : duration;
            var m = document.createElement('div');
            m.innerHTML = msg;
            m.style.cssText = `width:50%; min-width:180px; background:#000; opacity:0.5; height:auto;min-height: 30px; color:#fff; line-height:30px; text-align:center; border-radius:4px; position:fixed; top:${top}%; left:20%; z-index:999999;`
            document.body.appendChild(m);

            let that = this;
            setTimeout(function() {
                var d = 0.5;
                m.style.webkitTransition = '-webkit-transform ' + d + 's ease-in, opacity ' + d + 's ease-in';
                m.style.opacity = '0';

                setTimeout(function() {
                        document.body.removeChild(m);
                        that.toastList.splice(0, 1);
                    }, d * 1000);
            }, duration);
        },


        onClientStatusChanged(state) {
            if (state == ClientState.LOGINING) {
                this.clientStateText = '登录中';
            } else if (state == ClientState.LOGINED) {
                this.clientStateText = '已登录';
            } else if (state == ClientState.LOGOUTING) {
                this.clientStateText = '登出中';
            } else {
                this.clientStateText = '未登录';
            }
        },

        onRoomStatusChanged(roomState) {
            if (roomState == RoomState.IDLE) {
                this.roomStatus = 'idle';
                this.roomStateText = '空闲';
            } else if (roomState == RoomState.JOINING) {
                this.roomStatus = "joining";
                this.roomStateText = '正在加入房间...';
            } else if (roomState == RoomState.JOINED) {
                this.roomStatus = "joined";
                this.roomStateText = '已经加入房间';
            } else if (roomState == RoomState.LEAVING) {
                this.roomStatus = "leaving";
                this.roomStateText = '正在离开房间...';
            }
        },

        isScreenShareParticipant(participant) {
            if (participant.userId === this.room.getCallId()) {
             return true;
            }
            return false;
        },

        clearStreamNode() {
            const remote = document.getElementById("remote-video-container");
            if (remote != null) {
              this.clearAllNode(remote);
            }
            
            const local = document.getElementById("local-video-container");
            if (local != null) {
              this.clearAllNode(local);
            }
        },
    
        clearAllNode(parentNode){
            parentNode.innerHTML = '';
        },

        appendErrorContent(content) {
            let div =  this.createDivDOM();
            div.textContent = content;
            let viewId = 'error-code-detail'
            const box = document.querySelector(`#${viewId}`);
            box.appendChild(div);
        },

        createDivDOM() {
            const divDom = document.createElement('div');
            divDom.style.display = 'flex';
            divDom.style.margin = '0 auto';
            divDom.style.fontSize = '10px';
            return divDom;
        },

        appendLocalVideo(vid) {

            let userId = this.client.getUserId();

            let viewId = 'local-video-container'
            let local_video_container = document.querySelector(`#${viewId}`);
            const figure = document.createElement('figure');

            let div = this.createDivDOM();
            div.style.width = '180px';
            div.style.height = '320px';
            div.id = vid;
            div.style.backgroundColor = 'rgba(0,0,0)';
            figure.appendChild(div);

            let textDiv = this.createDivDOM();
            textDiv.style.width = '180px';
            textDiv.textContent = 'self';
            textDiv.style.display = 'flex';
            textDiv.style.justifyContent = 'center';
            figure.appendChild(textDiv);

            local_video_container.appendChild(figure);
        },

        appendRemoteVideo(userId, isScreenShare) {

            let viewId = 'remote-video-container'
            let remote_video_container = document.querySelector(`#${viewId}`);
            const figure = document.createElement('figure');
            
            figure.className = 'stream-wrapper';
            figure.id = 'figure-'+userId;

            let div = this.createDivDOM();
            div.style.width = '180px';
            div.style.height = '320px';
            div.id = 'video-' + userId;
            div.style.backgroundColor = 'rgba(0,0,0)';
            figure.appendChild(div);

            if (!isScreenShare) {
                const progress = document.createElement('progress');
                progress.style.display = 'flex';
                progress.style.margin = '0 auto';
                this.volumeMap['progress'+userId] = progress;
                figure.appendChild(progress);
            }

            let textDiv = this.createDivDOM();
            textDiv.style.width = '180px';
            textDiv.textContent = userId;
            textDiv.style.display = 'flex';
            textDiv.style.justifyContent = 'center';
            figure.appendChild(textDiv);

            remote_video_container.appendChild(figure);
        },

        removeRemoteVideo(userId) {
            let viewId = 'remote-video-container';
            let remote_video_container = document.querySelector(`#${viewId}`);

            let videoViewId = 'figure-' + userId;
            let video = document.getElementById(`${videoViewId}`);
            if (video) {
                remote_video_container.removeChild(video);
            }    
        },

        showRemoteVideo(show) {
            let viewId = 'remote-video-container';
            let remote_video_container = document.querySelector(`#${viewId}`);
            if (show) {
                remote_video_container.style.display = 'block';
            } else {
                remote_video_container.style.display = 'none';
            }
        },

        openFAQPage() {
            window.open(`../FAQ.html`);
        },

        strSize(str, charset) {
            let total = 0;
            charset = charset?.toLowerCase() || '';
            for (i = 0; i < str.length; i++) {
                let charCode = str.charCodeAt(i);
                if (charset === 'utf-16' || charset === 'utf16') {
                    total += charCode <= 0xffff ? 2 : 4;
                } else {
                    if (charCode <= 0x007f) {
                        total += 1;
                    } else if (charCode <= 0x07ff) {
                        total += 2;
                    } else if (charCode <= 0xffff) {
                        total += 3;
                    } else {
                        total += 4;
                    }
                }
            }
            return total;
        }
    }
  }).mount('#app');